'''
Created on Sep 11, 2015

@package: app
@author: scorp
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''

import hashlib
import re
import app.Utils as Utils  # pylint: disable=F0401
import dc_processor.Constants as CONSTS

logger = Utils.MPLogger().getLogger()


class ContentHashCalculator(object):

  ALGO_INCOME_BUF = 1
  ALGO_SIMPLE_SPLITTING = 2
  ALGO_SNOWBALL_SPLITTING = 3
  ALGO_SOUNDEX_SPLITTING = 4

  RE_SPLITTER = r'\s'
  
  # support synthetic fields constants
  SYNTHETIC_FIELD_DOMAIN_SUFFIX = '_domain'
  ACCUMULATED_TAGS_DELIMITER = ' '
  ORIGINAL_FIELDS_NAMES = [CONSTS.TAG_SOURCE_URL, CONSTS.TAG_LINK, CONSTS.TAG_FEED_URL, CONSTS.TAG_MEDIA]


  # # get synthetic field names from original
  #
  # @param originalFieldNames - original field names
  # @return synthetic field names
  @staticmethod
  def getSyntheticFieldNames(originalFieldNames=ORIGINAL_FIELDS_NAMES):
    # variable for result
    ret = []
    if isinstance(originalFieldNames, list):
      for fieldName in originalFieldNames:
        if isinstance(fieldName, basestring):
          ret.append(fieldName + ContentHashCalculator.SYNTHETIC_FIELD_DOMAIN_SUFFIX)

    return ret


  # # make accumulated buffer
  #
  # @param tagsDict - tags values dictionary
  # @param listHashedTags - list of tag's names for hash calculated contents
  # @return accumulated buffer
  @staticmethod
  def makeAccumulatedBuffer(tagsDict, listHashedTags):
    # variable for result
    ret = ""
    if isinstance(tagsDict, dict):
      syntheticFieldNames = ContentHashCalculator.getSyntheticFieldNames()
      for key, value in tagsDict.items():
        # accumulate to buffer usually tags
        if key in listHashedTags:
          for elem in ContentHashCalculator.extractValuesFromData(data=value):
            ret += (elem + ContentHashCalculator.ACCUMULATED_TAGS_DELIMITER)
            logger.debug("Tag '%s' was added to hash calculation buffer", str(key))

        # accumulate to buffer synthetic tags
        for syntheticFieldName in syntheticFieldNames:
          if syntheticFieldName in listHashedTags and \
            re.search(pattern=key, string=syntheticFieldName, flags=re.U + re.I) is not None:
            for elem in ContentHashCalculator.extractValuesFromData(data=value):
              # extract domain
              domain = Utils.parseHost(elem, True)
#               logger.debug("Extracted domain: %s", domain)
              if domain is not None:
                ret += (domain + ContentHashCalculator.ACCUMULATED_TAGS_DELIMITER)
                logger.debug("Synthetic tag '%s' with value '%s' was added to hash calculation buffer.", str(syntheticFieldName), str(domain))

    return ret


  # #extract values from data
  #
  # @param data - data value
  # @return extracted values list
  @staticmethod
  def extractValuesFromData(data):
    # variable for result
    values = []

    if isinstance(data, basestring):
      values.append(data)
    elif isinstance(data, list):
      for valueElem in data:
        if isinstance(valueElem, basestring):
          values.append(valueElem)
      
    return values


  @staticmethod
  def langDetect(incomeBuf, convertToFullName=True):
    ret = None

    try:
      from langdetect import detect
      langSmallName = detect(incomeBuf).split('-')[0]
      if convertToFullName:
        import pycountry
        ret = pycountry.languages.get(iso639_1_code=langSmallName).name.lower()
      else:
        ret = langSmallName
    except Exception as ecxp:
      logger.debug(">>> Some snowball exception = " + str(ecxp))

    return ret


  @staticmethod
  def commonSplitMethod(incomeBuf, minWLen):
    ret = re.split(ContentHashCalculator.RE_SPLITTER, incomeBuf, flags=re.LOCALE)

    if len(ret) > 0:
      ret = list(set(ret))
      ret.sort()
      ret = [x for x in ret if len(x) >= minWLen]

    return ret


  @staticmethod
  def hashCalculateSimple(incomeBuf, minWLen):
    ret = None

    splittedList = ContentHashCalculator.commonSplitMethod(incomeBuf, minWLen)
    if len(splittedList) > 0:
      ret = hashlib.md5(''.join(splittedList)).hexdigest()

    return ret


  @staticmethod
  def hashCalculateSnowball(incomeBuf, minWLen, additionData, stemmerLang):  # pylint: disable=W0613
    ret = None
    try:
      import snowballstemmer
      if stemmerLang is None:
        stemmerLang = ContentHashCalculator.langDetect(incomeBuf)
        if stemmerLang is None:
          stemmerLang = 'english'

      # if additionData is not None and type(additionData) in types.StringTypes:
      splittedList = ContentHashCalculator.commonSplitMethod(incomeBuf, minWLen)
      if len(splittedList) > 0:
        stemmer = snowballstemmer.stemmer(stemmerLang)
        for i in xrange(0, len(splittedList)):
          try:
            splittedList[i] = stemmer.stemWord(splittedList)
          except Exception as ecxp:
            splittedList[i] = ""
        ret = hashlib.md5(''.join(splittedList)).hexdigest()
    except Exception as ecxp:
      logger.debug(">>> Some snowball exception = " + str(ecxp))

    return ret


  @staticmethod
  def hashCalculateSoundex(incomeBuf, minWLen):
    ret = None
    try:
      import soundex
      splittedList = ContentHashCalculator.commonSplitMethod(incomeBuf, minWLen)
      if len(splittedList) > 0:
        s = soundex.getInstance()
        for i in xrange(0, len(splittedList)):
          try:
            splittedList[i] = s.soundex(splittedList[i])
          except Exception as ecxp:
            splittedList[i] = ""
        ret = hashlib.md5(''.join(splittedList)).hexdigest()
    except Exception as ecxp:
      logger.debug(">>> Some soundex exception = " + str(ecxp))

    return ret


  @staticmethod
  def hashCalculate(incomeBuf, algo, minWLen=3, additionData=None, stemmerLang=None):
    ret = None
    incomeBuf = incomeBuf.lower()
    if algo == ContentHashCalculator.ALGO_INCOME_BUF:
      ret = hashlib.md5(incomeBuf).hexdigest()
    elif algo == ContentHashCalculator.ALGO_SIMPLE_SPLITTING:
      ret = ContentHashCalculator.hashCalculateSimple(incomeBuf, minWLen)
    elif algo == ContentHashCalculator.ALGO_SNOWBALL_SPLITTING:
      ret = ContentHashCalculator.hashCalculateSnowball(incomeBuf, minWLen, additionData, stemmerLang)
    elif algo == ContentHashCalculator.ALGO_SOUNDEX_SPLITTING:
      ret = ContentHashCalculator.hashCalculateSoundex(incomeBuf, minWLen)

    return ret
