# coding: utf-8
"""
@package: dc
@file CollectURLs.py
@author Scorp <developers.hce@gmail.com>, alexv
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import copy
import json
import hashlib
import time
import datetime
import re
import urlparse
import lxml.html

try:
  import feedparser
except ImportError:
  feedparser = None

# from collections import OrderedDict

import app.Consts as APP_CONSTS
from app.Filters import Filters
from app.UrlNormalize import UrlNormalize
from app.Utils import UrlNormalizator
from app.UrlsLogging import UrlsLogging
from app.Utils import varDump
import app.Utils as Utils  # pylint: disable=F0401
import dc_processor.Constants as PCONSTS
import dc.EventObjects
from dc_crawler.Fetcher import SeleniumFetcher
import dc_crawler.Constants as CRAWLER_CONSTS
from dc_crawler.HTTPRedirectExtractor import HTTPRedirectExtractor
from dc_crawler.HTTPRedirectResolver import HTTPRedirectResolver
from dc_crawler.HTTPCookieResolver import HTTPCookieResolver
from dc_crawler.CollectUrlsLimits import CollectUrlsLimits
from dc_crawler.UrlLimits import UrlLimits
from dc_crawler.UrlChain import UrlChain
from dc_crawler.MimeTypeExtractor import MimeTypeExtractor
from dc_crawler.CollectURLsXPath import CollectURLsXPath

logger = Utils.MPLogger().getLogger()


class CollectURLs(object):

  BINARY_CONTENT_TYPE_PATTERN = re.compile('(text)|(xml)', re.I)
  COLLECT_POST_DATA_NAME = "COLLECT_POST_DATA"
  COLLECT_POST_DATA = "1"
  DC_URLS_TABLE_PREFIX = "urls_"
  PATTERN_WITH_PROTOCOL = re.compile('[a-zA-Z]+:(//)?')


  def __init__(self, isAbortedByTTL=None):
    self.crawledResource = None
    self.url = None
    self.dom = None
    self.realUrl = None
    self.baseUrl = None
    self.processorName = None
    self.batchItem = None
    self.feedItems = []
    self.feed = None
    self.siteProperties = None
    self.site = None
    self.dbWrapper = None
    self.autoRemoveProps = None
    self.processContentTypes = []
    self.postForms = None
    self.urlProcess = None
    self.robotsParser = None
    self.urlsXpathList = []
    self.isAbortedByTTL = (lambda: False) if isAbortedByTTL is None else isAbortedByTTL
    self.normMask = UrlNormalizator.NORM_DEFAULT


  # #checkFieldsIsNone method checks all class's mandatory fields
  #
  def checkFieldsIsNone(self):
    excludeList = ['feedItems', 'feed', 'processorName', 'autoDetectMime', 'processContentTypes',
                   'postForms', 'robotsParser', 'dom', 'dbWrapper', 'baseUrl']
    for field in self.__dict__:
      if field not in excludeList and (not hasattr(self, field) or getattr(self, field) is None):
        msg = "Mandatory field must be initialized, field Name = " + field
        logger.error(msg)
        raise Exception(msg)


  # # process method
  #
  # @param httpCode - http code
  # @param readOnly - boolean flag read only
  # @param httpApplyHeaders - http headers for apply
  def process(self, httpCode, readOnly=False, httpApplyHeaders=None):

    self.checkFieldsIsNone()
    if self.siteProperties is None:
      self.siteProperties = {}
    if self.processContentTypes is None:
      self.processContentTypes = []
    localSiteId = self.batchItem.siteId if self.batchItem.siteId else "0"
    nextStep = True
    useChains = False
    internalLinks, externalLinks = [], []
    maxURLsFromPage = 0
    params = []
    chainUrls = []
    isValidDomStructure = True
    formUrls = None
    formMethods = None
    formFields = None
    urlSet = []  # OrderedDict()  # set()
    rawUrlSet = []

    logger.debug("!!! self.site.maxURLsFromPage = " + str(self.site.maxURLsFromPage))
    logger.debug("!!! self.url.maxURLsFromPage = " + str(self.url.maxURLsFromPage))

    if self.site is not None and self.site.maxURLsFromPage is not None:
      maxURLsFromPage = self.site.maxURLsFromPage

    if self.url is not None and self.url.maxURLsFromPage is not None and self.url.maxURLsFromPage > 0:
      maxURLsFromPage = self.url.maxURLsFromPage

    if nextStep and self.crawledResource is not None and \
    not self.BINARY_CONTENT_TYPE_PATTERN.search(self.crawledResource.content_type):
      nextStep = False

    # don't parse url for 4XX or 5XX response
    if nextStep and self.crawledResource is not None:
      code_type = int(self.crawledResource.http_code) / 100
      if code_type == 4 or code_type == 5:
        nextStep = False

    if nextStep and self.crawledResource is not None and not self.crawledResource.html_content:
      nextStep = False

#     if nextStep and self.dom is None:
#      logger.debug("DOM is None")
#      nextStep = False

    if nextStep:
      useChains = True
      if self.dom is not None:
        isValidDomStructure = self.processProcessor(urlSet, self.dom, self.batchItem.urlObj, httpApplyHeaders)
        formUrls, formMethods, formFields = self.extractFormURL(self.dom, self.siteProperties)
#         urlSet.update(formUrls)
        urlSet.extend(formUrls)
      else:
        logger.debug("DOM is None")

      if self.url.type == dc.EventObjects.URL.TYPE_SINGLE:
        logger.debug("URL type: single")
        nextStep = False

      if nextStep and self.crawledResource.dynamic_fetcher_result_type == \
        SeleniumFetcher.MACRO_RESULT_TYPE_URLS_LIST:
        ul = None
        try:
          ul = json.loads(self.crawledResource.html_content)
        except Exception, err:
          logger.error("Error deserialize macro data from result string: %s\n%s", str(err), self.crawledResource.html_content)

        if ul is not None:
          logger.debug("Fill urlSet from macro results: %s items", str(len(ul)))
          if isinstance(ul, list):
#             urlSet.update([u for u in ul if isinstance(u, basestring) and u != ''])
            urlSet.extend([u for u in ul if isinstance(u, basestring) and u != ''])

    if nextStep:
      # (3) END
      urlTable = self.DC_URLS_TABLE_PREFIX + localSiteId
      self.urlProcess.urlTable = urlTable
      try:
        if self.siteProperties is not None and "RSS_FEED_ZERO_ITEM" in self.siteProperties and \
        int(self.siteProperties["RSS_FEED_ZERO_ITEM"]) == 1:
          if self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS:
            self.feedElementsProcessing(self.url.urlMd5, httpCode, self.url.url, localSiteId, self.url, self.url.url, params, maxURLsFromPage, True)
      except ValueError:
        logger.debug(">>> Wrong \"RSS_FEED_ZERO_ITEM\" property's value")

      # remove duplicate of candidates
      tmpUrlSet = []
      for x in urlSet:
        if x not in tmpUrlSet:
          tmpUrlSet.append(x)
      urlSet = tmpUrlSet

      rawUrlSet = copy.deepcopy(urlSet)
      logger.debug("URLs candidates collected %s items:\n%s", str(len(urlSet)), str(urlSet))
      if APP_CONSTS.URLS_LOGGING in self.siteProperties and self.siteProperties[APP_CONSTS.URLS_LOGGING] != "":
        for url in urlSet:
          UrlsLogging.addUrl(properties=self.siteProperties[APP_CONSTS.URLS_LOGGING],
                             url=url,
                             siteId=localSiteId,
                             stage=UrlsLogging.STAGE_COLLECT,
                             dbWrapper=self.dbWrapper,
                             log=logger)

      if self.site.maxURLs > 0 and len(urlSet) >= self.site.maxURLs:
        urlSet = urlSet[:self.site.maxURLs]
        logger.debug("Site maxURLs = %s limit reached.", str(self.site.maxURLs))

      if self.site.maxResources > 0 and len(urlSet) >= self.site.maxResources:
        urlSet = urlSet[:self.site.maxResources]
        logger.debug("Site maxResources = %s limit reached.", str(self.site.maxResources))

      countCnt = 0
      countErrors = 0
      for elemUrl in urlSet:
        if self.isAbortedByTTL():
          logger.debug("Aborted by TTL. All elements skipped.")
          break

        if elemUrl is None:
          logger.debug("Some url from urlSet is None, skipped.")
          continue

        elemUrl = elemUrl.strip()
        if elemUrl == '':
          logger.debug("Some url from urlSet is empty, skipped!")
          continue

        logger.debug("Before URL normalization url: %s", str(elemUrl))
        normalizedUrl = UrlNormalize.execute(siteProperties=self.siteProperties, base=self.baseUrl, url=elemUrl, supportProtocols=None, log=logger)
        logger.debug(" After URL normalization url: %s", str(normalizedUrl))
        if normalizedUrl is not None:
          elemUrl = normalizedUrl

        # Apply filter for URL
        if not self.filtersApply(inputFilters=self.site.filters, subject=elemUrl, depth=self.url.depth, wrapper=self.dbWrapper, siteId=localSiteId,
                                 currentBatchIterations=self.urlProcess.currentBatchIterations, siteProperties=self.siteProperties, url=elemUrl):
          logger.debug("Candidate URL not matched filters, skipped: %s", str(elemUrl))
          continue
        else:
          logger.debug("Candidate URL matched filters: %s", str(elemUrl))

        # Check type of the URL
        if self.url.type == dc.EventObjects.URL.TYPE_SINGLE:
          logger.debug("URL type is TYPE_SINGLE (not collect URL's). URL Skipped.")
          continue

        # Check URL's protocol
        if Filters(None, self.dbWrapper, localSiteId, 0, None, Filters.OC_RE, Filters.STAGE_COLLECT_URLS_PROTOCOLS).\
          isExistStage(Filters.STAGE_COLLECT_URLS_PROTOCOLS):
          logger.debug("Found filter for check 'COLLECT_URLS_PROTOCOLS'")
          # Apply filter for URL's protocol
          if not self.filtersApply(inputFilters=self.site.filters, subject=elemUrl, depth=self.url.depth, wrapper=self.dbWrapper, siteId=localSiteId,
                                   stage=Filters.STAGE_COLLECT_URLS_PROTOCOLS, currentBatchIterations=self.urlProcess.currentBatchIterations,
                                   siteProperties=self.siteProperties, url=elemUrl):
            logger.debug("Candidate URL not matched URL's protocols filters, skipped: %s", str(elemUrl))
            continue
          else:
            logger.debug("Candidate URL matched URL's protocols filters: %s", str(elemUrl))

        else:
          # Check from constants list if filter not set
          logger.debug("Not found filter for check 'COLLECT_URLS_PROTOCOLS'. Use constants list: %s", str(self.urlProcess.DEFAULT_PROTOCOLS))

          protocol = urlparse.urlsplit(elemUrl).scheme
          if protocol in self.urlProcess.DEFAULT_PROTOCOLS:
            logger.debug("Candidate URL has allowed protocol '%s': %s", str(protocol), str(elemUrl))
          else:
            logger.debug("Candidate URL has not allowed protocol '%s', skipped: %s", str(protocol), str(elemUrl))
            continue

#         logger.info("!!!      elemUrl: %s", str(elemUrl))
#         logger.info("!!! self.realUrl: %s", str(self.realUrl))
        # Fill internal and external link lists
        if Utils.parseHost(elemUrl) == Utils.parseHost(self.realUrl):
          internalLinks.append(elemUrl)
          logger.debug("URL classified as internal: %s", str(elemUrl))
        elif Utils.parseHost(elemUrl):
          externalLinks.append(elemUrl)
          logger.debug("URL classified as external: %s", str(elemUrl))
        else:  # not valid URL like http://
          logger.error("Found wrong URL: %s", str(elemUrl))

        localUrl = elemUrl
        self.urlProcess.urlObj = self.url
        self.urlProcess.url = elemUrl
        self.urlProcess.dbWrapper = self.dbWrapper

        # Check exist of the Url
        urlMd5 = hashlib.md5(elemUrl).hexdigest()
        self.urlProcess.url = elemUrl
        self.urlProcess.siteId = localSiteId
        self.urlProcess.urlTable = urlTable
        if self.urlProcess.isUrlExist(self.site.recrawlPeriod, urlMd5):
          logger.debug("Candidate URL %s already exist, skipped.", str(urlMd5))
          continue

        if self.site.maxURLs > 0:
          if httpCode == CRAWLER_CONSTS.HTTP_CODE_200:
            countCnt += 1
          else:
            countErrors += 1

          if self.dbWrapper is not None:
            currentCnt = self.urlProcess.readCurrentCnt(self.site.maxURLs)
            if currentCnt >= self.site.maxURLs or countCnt >= self.site.maxURLs or \
              (countCnt + countErrors) >= self.site.maxURLs:
              logger.debug("Site MaxURLs: %s limit is reached. countCnt = %s, currentCnt = %s",
                           str(self.site.maxURLs), str(countCnt), str(currentCnt))
              autoremovedURLs = self.urlProcess.autoRemoveURL(self.autoRemoveProps, self.site.recrawlPeriod, urlTable, self.dbWrapper)
              if autoremovedURLs == 0:
                logger.debug("No one URL auto removed, candidate URL skipped!")
                continue
              else:
                logger.debug("%s URLs auto removed.", str(autoremovedURLs))

            if currentCnt >= self.site.maxResources or countCnt >= self.site.maxResources or (countCnt + countErrors) >= self.site.maxResources:
              logger.debug("Site maxResources = %s limit is reached. countCnt = %s, currentCnt = %s",
                           str(self.site.maxResources), str(countCnt), str(currentCnt))
              autoremovedURLs = self.urlProcess.autoRemoveURL(self.autoRemoveProps, self.site.recrawlPeriod, urlTable, self.dbWrapper)
              if autoremovedURLs == 0:
                logger.debug("No one URL auto removed, candidate URL skipped!")
                continue
              else:
                logger.debug("%s URLs auto removed.", str(autoremovedURLs))

        # detect collected url mime type and ignore non-match URL
        # (7) Detect collected url mime type and ignore non-match URL
        logger.info("!!! Detect collected URL's mime type !!!")
        detectedMime = None
        # Check URL content-type by HEAD request
        if Filters(None, self.dbWrapper, localSiteId, 0, None, Filters.OC_RE, Filters.STAGE_COLLECT_URLS,
          Filters.SELECT_SUBJECT_HTTP_HEAD).isExistStage(Filters.STAGE_COLLECT_URLS):
          logger.debug("Found filter for check Content-Type on stage 'COLLECT_URLS'")
          # Apply filter for URL content-type
          if self.filtersApply(inputFilters=self.site.filters, subject=elemUrl, depth=self.url.depth, wrapper=self.dbWrapper, siteId=localSiteId,
                                   stage=Filters.STAGE_COLLECT_URLS, selectSubject=Filters.SELECT_SUBJECT_HTTP_HEAD, defaultValue=True,
                                   currentBatchIterations=self.urlProcess.currentBatchIterations, siteProperties=self.siteProperties, url=elemUrl):
            logger.debug("Candidate URL matched URL's pattern. Url: %s", str(elemUrl))
            detectedMime = MimeTypeExtractor.extractUrlMime(url=elemUrl, siteProperties=self.siteProperties, log=logger, default=None)
            # Check for allowed types list
            if isinstance(self.processContentTypes, list) and isinstance(detectedMime, basestring) and detectedMime not in self.processContentTypes:
              logger.debug("Candidate URL MIME type '%s' is not matched in allowed content types: %s, skipped url: %s", str(detectedMime),
                           str(self.processContentTypes), str(elemUrl))
              continue
            else:
              logger.debug("Candidate URL MIME type '%s' is allowed for content types: %s, url: %s", str(detectedMime), str(self.processContentTypes),
                           str(elemUrl))
        # (7) END

# # TODO check and fix
#
#         if "ROBOTS_COLLECT" not in self.siteProperties or int(self.siteProperties["ROBOTS_COLLECT"]) > 0:
#           logger.debug("Robots.txt obey mode is ON")
#           if self.robotsParser and self.robotsParser.loadRobots(elemUrl, self.batchItem.siteId, httpApplyHeaders, proxyName):
#             isAllowed, retUserAgent = self.robotsParser.checkUrlByRobots(elemUrl, self.batchItem.siteId, httpApplyHeaders)
#             if not isAllowed:
#               logger.debug("URL " + elemUrl + " is NOT Allowed by user-agent:" + str(retUserAgent))
#               self.urlProcess.updateURLForFailed(APP_CONSTS.ERROR_ROBOTS_NOT_ALLOW, self.batchItem)
#               continue

        self.urlProcess.siteId = localSiteId
        depth = self.urlProcess.getDepthFromUrl(self.batchItem.urlId)

#         logger.info("="*50)
#         logger.info("!!!>>>             elemUrl = %s", str(elemUrl))
#         logger.info("!!!>>> self.urlProcess.url = %s", str(self.urlProcess.url))
        if elemUrl is not None:
          # usually insert new URL
          localUrlObj = self.urlProcess.createUrlObjectForCollectURLs(siteProperties=self.siteProperties,
                                                                 urlObj=self.url,
                                                                 siteObj=self.urlProcess.site,
                                                                 siteId=localSiteId,
                                                                 url=elemUrl,
                                                                 normMask=self.normMask,
                                                                 formMethods=formMethods,
                                                                 parentMd5=self.batchItem.urlId,
                                                                 depth=depth,
                                                                 detectedMime=detectedMime,
                                                                 maxURLsFromPage=self.site.maxURLsFromPage)

          # update counters of external and internal links
          localUrlObj.linksI = len(internalLinks)
          localUrlObj.linksE = len(externalLinks)

#           logger.info("!!!>> self.processorName = %s", str(self.processorName))
#           logger.info("!localUrlObj.contentType = '%s'", str(localUrlObj.contentType))
          if self.processorName in [PCONSTS.PROCESSOR_FEED_PARSER , PCONSTS.PROCESSOR_RSS]:
            # and   localUrlObj.contentType != "" and localUrlObj.contentType != dc.EventObjects.URL.CONTENT_TYPE_TEXT_HTML:
            self.feedElementsProcessing(urlMd5, httpCode, elemUrl, localSiteId, localUrlObj, localUrl, params, maxURLsFromPage, False)
          else:
            logger.debug("!!! params.append(localUrlObj) %s", varDump(localUrlObj))
            params.append(localUrlObj)

#     logger.debug("!!! urlNew() params1: %s", varDump(params))

#     logger.debug("!!! self.siteProperties: %s", str(self.siteProperties))
    if useChains and APP_CONSTS.URL_CHAIN in self.siteProperties and self.siteProperties[APP_CONSTS.URL_CHAIN] is not None:
      # localChainDict = Utils.jsonLoadsSafe(jsonString=self.siteProperties[APP_CONSTS.URL_CHAIN], default={}, log=logger)
      # logger.debug("!!! (URL_CHAIN) localChainDict: %s", str(localChainDict))

      logger.debug("!!! URL_CHAIN !!!")
      logger.debug("!!! self.processorName = %s", str(self.processorName))

      urlChain = UrlChain(self.siteProperties[APP_CONSTS.URL_CHAIN], logger)
      logger.debug("!!! (URL_CHAIN) urlSet[0] = %s", str(urlSet[0] if len(urlSet) > 0 else None))

      mainUrl = urlSet[0] if len(urlSet) > 0 else None
      logger.debug("!!! (URL_CHAIN) mainUrl = %s", str(mainUrl))

      urlPattern = urlPatternExclude = None
      if mainUrl is not None:
        urlPattern = urlChain.getUrlPattern(mainUrl)
        logger.debug("!!! (URL_CHAIN) urlPattern = %s", str(urlPattern))

        urlPatternExclude = urlChain.getUrlPatternExclude(mainUrl)
        logger.debug("!!! (URL_CHAIN) urlPatternExclude = %s", str(urlPatternExclude))

      depth = self.urlProcess.getDepthFromUrl(self.batchItem.urlId)
      logger.debug("!!! (URL_CHAIN) depth = %s", str(depth))

      if urlPattern is not None and urlPattern != "":

        for elemUrl in urlSet:
          logger.debug("!!! CHAIN elemUrl: %s", str(elemUrl))

          if elemUrl is None:
            logger.debug("Some CHAIN url from urlSet is None")
            continue

          # normalization before URL redirect resolve
          normalizedUrl = Utils.urlNormalization(base=self.baseUrl, url=elemUrl, supportProtocols=None, log=logger)
          if normalizedUrl is not None:
            elemUrl = normalizedUrl

          # Apply filter to Url
          if not self.filtersApply(inputFilters=self.site.filters, subject=elemUrl, depth=self.url.depth, wrapper=self.dbWrapper, siteId=localSiteId,
                                   currentBatchIterations=self.urlProcess.currentBatchIterations, siteProperties=self.siteProperties, url=elemUrl):
            logger.debug("Candidate URL CHAIN not matched filters, skipped: %s", str(elemUrl))
            continue
          else:
            logger.debug("Candidate URL CHAIN matched filters: %s", str(elemUrl))

          # Apply url pattern
          if not urlChain.urlPatternApply(urlPattern, elemUrl):
            logger.debug("Candidate URL CHAIN not passed 'url_pattern'\npattern: %s,\nskipped url: %s", str(urlPattern), str(elemUrl))
            continue
          else:
            logger.debug("Candidate URL CHAIN passed 'url_pattern'\npattern: %s,\nurl: %s", str(urlPattern), str(elemUrl))

          # Apply url pattern exclude
          if not urlChain.urlPatternExcludeApply(urlPatternExclude, elemUrl):
            logger.debug("Candidate URL CHAIN not passed 'url_pattern_exclude'\npattern: %s,\nskipped url: %s", str(urlPatternExclude), str(elemUrl))
            continue
          else:
            logger.debug("Candidate URL CHAIN passed 'url_pattern_exclude'\npattern: %s,\nurl: %s", str(urlPatternExclude), str(elemUrl))

          # Check URL content-type by HEAD request
          logger.info("!!! Detect URL CHAIN mime type !!!")
          detectedMime = None
          if Filters(None, self.dbWrapper, localSiteId, 0, None, Filters.OC_RE, Filters.STAGE_COLLECT_URLS,
            Filters.SELECT_SUBJECT_HTTP_HEAD).isExistStage(Filters.STAGE_COLLECT_URLS):
            logger.debug("Found filter for check URL CHAIN Content-Type on stage 'COLLECT_URLS'")
            # Apply filter for URL content-type
            if self.filtersApply(inputFilters=self.site.filters, subject=elemUrl, depth=self.url.depth, wrapper=self.dbWrapper, siteId=localSiteId,
                                     stage=Filters.STAGE_COLLECT_URLS, selectSubject=Filters.SELECT_SUBJECT_HTTP_HEAD, defaultValue=True,
                                     currentBatchIterations=self.urlProcess.currentBatchIterations, siteProperties=self.siteProperties, url=elemUrl):
              logger.debug("Candidate URL matched URL's pattern. Url: %s", str(elemUrl))
              detectedMime = MimeTypeExtractor.extractUrlMime(url=elemUrl, siteProperties=self.siteProperties, log=logger, default=None)
              logger.debug("detectedMime = '%s'", str(detectedMime))
              if isinstance(self.processContentTypes, list) and isinstance(detectedMime, basestring) and detectedMime not in self.processContentTypes:
                logger.debug("Candidate URL MIME type '%s' is not matched in allowed content types: %s, skipped url: %s", str(detectedMime),
                             str(self.processContentTypes), str(elemUrl))
                continue
              else:
                logger.debug("Candidate URL MIME type '%s' is allowed for content types: %s, url: %s", str(detectedMime), str(self.processContentTypes),
                           str(elemUrl))

          urlMd5 = hashlib.md5(elemUrl).hexdigest()
          self.urlProcess.url = elemUrl
          self.urlProcess.siteId = localSiteId
          self.urlProcess.urlObj = self.url
          try:
            localUrlObj = self.urlProcess.createUrlObjForChain(urlMd5, formMethods, self.batchItem.urlId, depth, detectedMime, self.site.maxURLsFromPage)

            logger.debug("!!! (URL_CHAIN) localUrlObj: %s", varDump(localUrlObj))
            if localUrlObj is not None:
              chainUrls.append(copy.deepcopy(localUrlObj))
          except Exception as excp:
            logger.error("Error in URL_CHAIN deserialize, excp = " + str(excp))

#     logger.debug("!!! urlNew() params2: %s", varDump(params))

    if len(urlSet) > 0 and len(params) == 0:
      logger.debug("Zero urls are collected for len(urlSet): %s", str(len(urlSet)))
    elif len(params) > 0:
      if APP_CONSTS.URL_LIMITS in self.siteProperties:
        #logger.debug("!!! %s !!!", str(APP_CONSTS.URL_LIMITS))
        params = UrlLimits.applyLimits(properties=self.siteProperties[APP_CONSTS.URL_LIMITS], urlsList=params, log=logger)
      logger.debug("Collected and send to insert as new: %s", str(len(params)))

#     logger.debug("!!! urlNew() params3: %s", varDump(params))

    params1 = []
    for paramsItem in params:
      found = False
      for chainUrlsItem in chainUrls:
        if paramsItem.url == chainUrlsItem.url:
          found = True
          break
      if not found:
        params1.append(paramsItem)
    params = params1

    if not readOnly:
      if self.dbWrapper is not None:
        logger.debug("!!! urlNew() params4: %s", varDump(params))
        self.dbWrapper.urlNew(params)
        self.dbWrapper.addContentURLNew(params)

        logger.debug("!!! urlNew() chainUrls: %s", varDump(chainUrls))
        self.dbWrapper.urlNew(chainUrls)
        self.urlProcess.updateTypeForURLObjects(chainUrls, dc.EventObjects.URL.TYPE_CHAIN)

        self.dbWrapper.collectedURLsRecalculating(localSiteId)

        if APP_CONSTS.URLS_LOGGING in self.siteProperties and self.siteProperties[APP_CONSTS.URLS_LOGGING] != "":
          for elemUrlObj in params + chainUrls:
            UrlsLogging.addUrl(properties=self.siteProperties[APP_CONSTS.URLS_LOGGING],
                               url=elemUrlObj.url,
                               siteId=localSiteId,
                               stage=UrlsLogging.STAGE_INSERT,
                               dbWrapper=self.dbWrapper,
                               log=logger)

            if elemUrlObj not in rawUrlSet:
              UrlsLogging.addUrl(properties=self.siteProperties[APP_CONSTS.URLS_LOGGING],
                                 url=elemUrlObj.url,
                                 siteId=localSiteId,
                                 stage=UrlsLogging.STAGE_SKIPPED,
                                 dbWrapper=self.dbWrapper,
                                 log=logger)

    if formFields is not None and self.postForms is not None and self.dbWrapper is not None:
      fieldParams = self.getFieldParams(formFields, self.postForms, localSiteId)
      self.insertNewSiteProperties(fieldParams, self.dbWrapper, localSiteId)

#     logger.debug("Return from collectURLs:\n%s\n%s\n%s\n%s\n%s\n%s", str(nextStep), str(internalLinks),
#                  str(externalLinks), str(params), str(self.feedItems), str(chainUrls))

    return nextStep, internalLinks, externalLinks, params, self.feedItems, chainUrls, isValidDomStructure


  # # feedElementsProcessing processed rss element
  #
  # @param urlMd5 - element url's urlMd5
  # @param httpCode - http code of http response
  # @param elemUrl - element url's url
  # @param localSiteId - siteId
  # @param localUrlObj - element's urlObj
  # @param localUrl - root element's url
  # @param params - element container (List type)
  # @param maxURLsFromPage - max URLs from page
  # @param rootFeed - bool param that specifyied use feed element from rss struture of entities element
  # @return - None
  def feedElementsProcessing(self, urlMd5, httpCode, elemUrl, localSiteId, localUrlObj, localUrl, params, maxURLsFromPage, rootFeed=False):

    if APP_CONSTS.COLLECT_URLS_LIMITS in self.siteProperties:
      logger.debug("Property '%s': %s", str(APP_CONSTS.COLLECT_URLS_LIMITS), str(self.siteProperties[APP_CONSTS.COLLECT_URLS_LIMITS]))
      maxURLsFromPage = CollectUrlsLimits.execute(properties=self.siteProperties[APP_CONSTS.COLLECT_URLS_LIMITS],
                                                  url=elemUrl,
                                                  optionsName=CollectUrlsLimits.MAX_URLS_FROM_PAGE,
                                                  default=maxURLsFromPage,
                                                  log=logger)
      logger.debug("Result maxURLsFromPage = %s", str(maxURLsFromPage))

    if maxURLsFromPage > 0 and len(self.feedItems) >= maxURLsFromPage:
      logger.debug("Site maxURLsFromPage = %s limit reached on %s number.", str(maxURLsFromPage), str(len(self.feedItems)))
    else:
#       logger.info("!!!>> self.feed = %s", str(self.feed))

      if self.feed is not None:
        self.urlProcess.url = elemUrl
        self.urlProcess.siteId = localSiteId
        self.urlProcess.urlObj = localUrlObj
        self.urlProcess.dbWrapper = self.dbWrapper
#         localRet = self.urlProcess.fillRssFieldInUrlObj(localUrl, self.url.url, self.batchItem, self.processorName, self.feed, rootFeed)
        localRet = self.urlProcess.fillRssFieldInUrlObj(elemUrl, self.url.url, self.batchItem, self.processorName, self.feed, rootFeed)
#         logger.info("!!!>> localRet = %s", str(localRet))

        self.urlProcess.urlObj = None
        if localRet is not None:
          localRet["urlMd5"] = urlMd5
          if localRet["urlObj"] is not None:
            localRet["urlObj"].httpCode = httpCode
            localRet["urlObj"].processingDelay = 0
            localRet["urlObj"].parentMd5 = self.url.urlMd5
            params.append(localRet["urlObj"])
            self.feedItems.append(localRet)
      else:
        logger.debug("self.feed is None!")


  # # processProcessor processed element from url's dom
  #
  # @param urlSet - incoming url's list (links from page)
  # @param dom - page's dom model
  # @param urlObj - element of urlObject
  # @return boolean flag True if valid DOM structure or otherwise False
  def processProcessor(self, urlSet, dom, urlObj, headers=None):
#     logger.debug("!!! self.processorName = %s", str(self.processorName))
#     logger.debug("!!! urlObj.type = %s", str(urlObj.type))
#     logger.debug("!!! urlObj.url = %s", str(urlObj.url))

    # variable for result
    isValidDomStructure = True

    logger.debug("self.processContentTypes: %s", str(self.processContentTypes))

    if self.crawledResource.content_type.strip() in self.processContentTypes:
      if self.processorName == PCONSTS.PROCESSOR_RSS:
        try:
          if feedparser is None:
            raise Exception("Module 'feedparser'  not found.")

          self.feedItems = []
          # Add one more date parsing handler function to fix some wrong datetime format cases; added by bgv
          feedparser.registerDateHandler(self.feedparserParseDateFixes)

          # Remove handlers to process all tags as unknown to save their names unchanged
          if not (self.siteProperties is not None and "RSS_FEEDPARSER_MODE" in self.siteProperties and \
                  int(self.siteProperties["RSS_FEEDPARSER_MODE"]) > 0):
            import inspect
            # , "_start_guid"
            excludes = ["_start_rss", "_start_channel", "_start_feed", "_start_item", "_start_link",
                        "_start_admin_errorreportsto", "_start_admin_generatoragent", "_start_guid", "_start_id",
                        "_start_entry", "_start_enclosure"]
            for methodName, functionObject in inspect.getmembers(feedparser._FeedParserMixin, predicate=inspect.ismethod):  # pylint: disable=W0612,W0212,C0301
              if methodName.startswith("_start_") and methodName not in excludes:
                delattr(feedparser._FeedParserMixin, methodName)  # pylint: disable=W0212
                endMethodName = methodName.replace("_start_", "_end_")
                if hasattr(feedparser._FeedParserMixin, endMethodName):  # pylint: disable=W0212
                  delattr(feedparser._FeedParserMixin, endMethodName)  # pylint: disable=W0212
            setattr(feedparser._FeedParserMixin, "_normalize_attributes", self._normalize_attributes)  # pylint: disable=W0212
            feedparser.FeedParserDict.keymap["guid"] = "guid"
            logger.debug("Feedparser in modified mode")
          else:
            logger.debug("Feedparser in native mode")

          isValidDomStructure = Utils.validateDomStructure(self.crawledResource.html_content, urlObj.url, logger)

          self.feed = feedparser.parse(self.crawledResource.html_content)
          logger.debug("feedparser return 'version' = '%s'", str(self.feed['version']))

          if self.feed['version'] == "":
            isValidDomStructure = False

          logger.info("Loaded from feed %s elements.", str(len(self.feed.entries)))
#           logger.debug("!!! self.feed.entries = %s", json.dumps(self.feed.entries, ensure_ascii=False, indent=2))

          for i in range(len(self.feed.entries)):
            entryUrl = None
            if hasattr(self.feed.entries[i], PCONSTS.TAG_LINK):
              entryUrl = getattr(self.feed.entries[i], PCONSTS.TAG_LINK)

            if hasattr(self.feed.entries[i], PCONSTS.ORIGINAL_LINK_NAME):
              entryUrl = getattr(self.feed.entries[i], PCONSTS.ORIGINAL_LINK_NAME)

            if entryUrl is None:
              logger.info("Couldn't extract entry URL from: %s", varDump(self.feed.entries[i], stringifyType=0))
              continue

            # resolve HTTP redirect URL
            resUrl = HTTPRedirectExtractor.extract(siteProperties=self.siteProperties,
                                                   siteId=self.batchItem.siteId,
                                                   url=entryUrl,
                                                   headers=headers,
                                                   httpTimeout=self.url.httpTimeout,
                                                   httpMethod=self.url.httpMethod,
                                                   fetchType=self.site.fetchType,
                                                   dbWrapper=self.dbWrapper,
                                                   filters=self.site.filters,
                                                   log=logger,
                                                   stage=HTTPRedirectResolver.RedirectProperty.VALUE_STAGE_COLLECT_URL,
                                                   cookiesStage=HTTPCookieResolver.STAGE_RSS)
            # URL normalization
#             logger.debug("Before URL normalization resUrl = %s", str(resUrl))
            resUrl = UrlNormalize.execute(siteProperties=self.siteProperties, base=self.baseUrl, url=resUrl, supportProtocols=None, log=logger)
#             logger.debug("After URL normalization resUrl = %s", str(resUrl))
            if resUrl is None:
              urlSet.append(entryUrl)
            else:
              urlSet.append(resUrl)
              if isinstance(self.feed.entries[i], dict):
                self.feed.entries[i].update({PCONSTS.TAG_SOURCE_URL: resUrl})
                self.feed.entries[i].update({PCONSTS.TAG_BASE_URL: entryUrl})

        except Exception, err:
          logger.error("Parse RSS feed error: %s", str(err))

      # won't collect urls from rss feed resources
      else:
        urlSet = CollectURLsXPath.execute(siteProperties=self.siteProperties, urlSet=urlSet, dom=dom, urlsXpathListFromConfig=self.urlsXpathList,
                                          url=urlObj.url, log=logger)

    logger.debug("urlSet: %s", str(urlSet))

    return isValidDomStructure

#   # #Evaluate date macro
#   #
#   # @param localPattern - input pattern as string
#   # @param dateFromat - format for %@DATE... macro
#   # @return patterns list with date macro evaluated and replaced with values
#   def evaluateDateMacro(self, localPattern, dateFromat):
# #     logger.debug("Input localPattern = %s", str(localPattern))
#     try:
#       regex = re.compile("%@(DATE)\\(([\\+|\\-]\\d{1,2})\\)%")
#       matchArray = regex.findall(localPattern)
#       for i in matchArray:
#         t = time.strftime(dateFromat, time.gmtime(time.time() + datetime.timedelta(hours=(+int(i[1]))).seconds))
#         localPattern = localPattern.replace("%@" + i[0] + "(" + i[1] + ")%", t)
#
#         localPattern = Filters.getGmtTime(localPattern, logger)
#
#     except Exception, err:
#       logger.error(str(err))
#
# #     logger.debug("Output localPattern = %s", str(localPattern))
#
#     return localPattern


  # # extrace URL, form action, and fields from html dom
  #
  # @param dom the dom tree
  # @return form_urls    sequence of urls extracted
  # @return form_methods  dict of form methods, {form_action: form_method}
  # @return form_fields  dict of fields {field_name: field_value}
  def extractFormURL(self, dom, siteProperties):
    formUrls, formMethods, formFields = [], {}, {}
    if self.COLLECT_POST_DATA_NAME in siteProperties and \
    siteProperties['COLLECT_POST_DATA'] == self.COLLECT_POST_DATA:
      for form in dom.xpath("//form"):
        formAction = None
        formMethod = 'get'
        for attr in form.keys():
          if attr.lower() == "action":
            formAction = form.get(attr)
            formUrls.append(formAction)
          elif attr.lower() == "method":
            formMethod = form.get(attr)
        if not formAction:
          continue
        formMethods[formAction] = formMethod
        for field in form.getchildren():
          tagName, tagValue = None, ""
          for fieldTag in field.keys():
            if fieldTag.lower() == "name":
              tagName = field.get(fieldTag)
            elif fieldTag.lower() == "value":
              tagValue = field.get(fieldTag)
          if tagName:
            formFields[tagName] = tagValue
      logger.info("extracted form data, formUrls:%s, formMethods:%s, formFields:%s", \
      formUrls, formMethods, formFields)
    return formUrls, formMethods, formFields


  # # Applys filters and returns bool result
  #
  # @param inputFilters - input sites filters list
  # @param subject - subject for apply filter
  # @param depth - depth value
  # @param wrapper - DBTasksWrapper instance
  # @param siteId - site ID used with db-task wrapper
  # @param fields - dictionary values of support macro names ('PDATE' and other)
  # @param opCode - operation code
  # @param stage - stage of apply filter
  # @param selectSubject - select subject use select from DB
  # @param defaultValue - default value for result
  # @param batchItem - batch item instance
  # @param currentBatchIterations - current batch iterations
  # @param siteProperties - site properties dictionary
  # @param url - URL string for domain check
  # @return True  if filter is good or False otherwise
  @staticmethod
  def filtersApply(inputFilters, subject, depth, wrapper, siteId, fields=None, opCode=Filters.OC_RE, stage=Filters.STAGE_COLLECT_URLS, selectSubject=None,
                   defaultValue=False, batchItem=None, currentBatchIterations=None, siteProperties=None, url=None):
    ret = defaultValue
    fValue = Utils.generateReplacementDict()
    fValue.update({"MAX_DEPTH": str(depth)})

    if inputFilters is not None:
      for inputFilter in inputFilters:
        if inputFilter.stage == Filters.STAGE_ALL or inputFilter.stage == Filters.STAGE_REDIRECT_URL:
          inputFilter.stage = Filters.STAGE_COLLECT_URLS

#     logger.debug(">>> Filters() (2.1) fields: " + varDump(fields) + " inputFilters: " + varDump(inputFilters))
    localFilters = Filters(filters=inputFilters, dbTaskWrapper=wrapper, siteId=siteId, readMode=0, fields=fields, opCode=opCode, stage=stage,
                           selectSubject=selectSubject, batchItem=batchItem, currentBatchIterations=currentBatchIterations,
                           siteProperties=siteProperties, url=url)

    # logger.debug(">>> before filter include = " + subject[:255] + ' . . . ')
    fResult = localFilters.filterAll(stage, fValue, Filters.LOGIC_OR, subject, 1)
    logger.debug(">>> filter result include - " + str(fResult))
    for elem in fResult:
#       logger.debug('elem = ' + str(elem) + ' type: ' + str(type(elem)))
      if elem > 0:
        ret = True
        break

    if ret is True:
      # logger.debug(">>> before filter exclude = " + subject[:255] + ' . . . ')
      fResult = localFilters.filterAll(stage, fValue, Filters.LOGIC_OR, subject, -1)
      logger.debug(">>> filter result exclude - " + str(fResult))
      for elem in fResult:
#         logger.debug('elem = ' + str(elem) + ' type: ' + str(type(elem)))
        if elem > 0:
          ret = False
          break

#     logger.debug("Verdict: " + str(ret))
    return ret


  # #getFieldParams method fill and returns post forms
  #
  # formFields - post fields dict
  # postForms - post form list
  # siteId - site's id
  def getFieldParams(self, formFields, postForms, siteId):
    ret = []
    for fieldName, fieldValue in formFields.iteritems():
      if fieldName in postForms:
        continue
      logger.debug("field_name: %s", fieldName)
      logger.debug("field_value: %s", fieldValue)
      ret.append((siteId, "HTTP_POST_FORM_" + fieldName, fieldValue))
    return ret


  # #insertNewSiteProperties update Site (add new properties in correspond table)
  #
  # params - params list
  # wrapper - db-task wrapper
  # siteId - site's id
  def insertNewSiteProperties(self, params, wrapper, siteId):
    if siteId is not None and hasattr(params, '__iter__') and len(params) > 0:
      localSiteUpdate = dc.EventObjects.SiteUpdate(siteId)
      for attr in localSiteUpdate.__dict__:
        if hasattr(localSiteUpdate, attr):
          setattr(localSiteUpdate, attr, None)
      localSiteUpdate.updateType = dc.EventObjects.SiteUpdate.UPDATE_TYPE_APPEND
      localSiteUpdate.id = siteId
      localSiteUpdate.properties = []
      for param in params:
        newPropElem = {}
        newPropElem["siteId"] = param[0]
        newPropElem["name"] = param[1]
        newPropElem["value"] = param[2]
        localSiteUpdate.properties.append(newPropElem)
      wrapper.siteNewOrUpdate(localSiteUpdate, stype=dc.EventObjects.SiteUpdate)


  # #feedparserParseDateFixes method to fix date parsing for the feedparser
  #
  # @param aDateString the date string to parse
  def feedparserParseDateFixes(self, aDateString):
    ret = None
    ds = aDateString

    # Assumes that date format broken and contains the semicolon ":" in TZ like: "Wed, 19 Aug 2015 08:45:53 +01:00"
    parts = ds.split(' ')
    if ("+" in parts[len(parts) - 1] or "-" in parts[len(parts) - 1]) and ":" in parts[len(parts) - 1]:
      parts[len(parts) - 1] = parts[len(parts) - 1].replace(":", "")
      ds = " ".join(parts)
      # ret = feedparser._parse_date_rfc822(ds)
      ret = feedparser._parse_date(ds)  # pylint: disable=W0212

    return ret


  def _normalize_attributes(self, kv):
    return (kv[0], kv[1])
