# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
CrawlerDataStorage Class content main handler for crawler data storage functional.

@package: dc_crawler
@file CrawlerDataStorage.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import os
import glob
import base64
import datetime
import collections

from dc_crawler.DBTasksWrapper import DBTasksWrapper
from dc_crawler.HTTPHeaders import HTTPHeaders
from app.Utils import PathMaker
import dc.EventObjects as dc_event
import dc.Constants as DC_CONSTANTS


class CrawlerDataStorage(object):
  # # Constants used in class
  CONTENT_DATA_NAME = 'data'
  DATETIME_SUFFIX_FORMAT = "_%Y_%m_%d_%H_%M_%S"

  DEFAULT_VALUE_KEEP_OLD_RESOURCE = False

  # # Error messages used in class
  ERROR_INPUT_DATA = "Wrong input data '%s' = %s"


  # # Initialization
  #
  # @param dbWrapper - DBTasksWrapper instance
  # @param log - logger instance
  # @param rawDataDir - raw data directory
  # @param keepOldResource - boolean flag keep old resources
  def __init__(self, dbWrapper=None, log=None, rawDataDir=None, keepOldResource=None):
    self.dbWrapper = dbWrapper
    self.logger = log
    self.rawDataDir = rawDataDir
    self.keepOldResource = self.DEFAULT_VALUE_KEEP_OLD_RESOURCE if keepOldResource is None else keepOldResource
    self.urlPutList = []


  # # Get header content
  #
  # @param rawDataDir - raw data directory
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param log - logger instance
  # @return extracted header content
  @staticmethod
  def getHeaderContent(rawDataDir, siteId, urlMd5, log=None):
    # variable for result
    headerContent = None
    try:
      if not isinstance(rawDataDir, basestring) or rawDataDir == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('rawDataDir', str(rawDataDir)))

      if not isinstance(siteId, basestring) or siteId == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('siteId', str(siteId)))

      if not isinstance(urlMd5, basestring) or urlMd5 == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('urlMd5', str(urlMd5)))

      if not os.path.isdir(rawDataDir):
        os.makedirs(rawDataDir)

      dataDir = (rawDataDir if rawDataDir[-1] == os.path.sep else rawDataDir + os.path.sep) + siteId + os.path.sep + PathMaker(urlMd5).getDir()
      fileMask = (dataDir + os.path.sep + "*" + DC_CONSTANTS.RAW_DATA_HEADERS_SUFF)
      files = sorted(glob.glob(fileMask), key=os.path.getctime, reverse=False)

      if len(files) > 0:
        if log is not None:
          log.debug("Open file: %s", str(files[0]))

        with open(files[0], 'r') as f:
          headerContent = f.read()

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return headerContent


  # # Get file suffix
  #
  # @param contentType - type of the content
  # @return resolved suffix
  @staticmethod
  def getFileSuffix(contentType):
    ret = ""
    if contentType == dc_event.Content.CONTENT_RAW_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_SUFF
    elif contentType == dc_event.Content.CONTENT_TIDY_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_TIDY_SUFF
    elif contentType == dc_event.Content.CONTENT_HEADERS_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_HEADERS_SUFF
    elif contentType == dc_event.Content.CONTENT_REQUESTS_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_REQESTS_SUFF
    elif contentType == dc_event.Content.CONTENT_META_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_META_SUFF
    elif contentType == dc_event.Content.CONTENT_COOKIES_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_COOKIES_SUFF
    elif contentType == dc_event.Content.CONTENT_DYNAMIC_CONTENT:
      ret = DC_CONSTANTS.RAW_DATA_DYNAMIC_SUFF
    elif contentType == dc_event.Content.CONTENT_CHAIN_PARTS:
      ret = DC_CONSTANTS.RAW_DATA_CHAIN_SUFF

    return ret


  # # Rename resource file
  #
  # @param rawDataDir - raw data directory
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param contentType - type of the content
  # @param log - logger instance
  # @return - None
  @staticmethod
  def renameResourceFile(rawDataDir, siteId, urlMd5, contentType, log=None):
    try:
      if not isinstance(rawDataDir, basestring) or rawDataDir == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('rawDataDir', str(rawDataDir)))

      if not isinstance(siteId, basestring) or siteId == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('siteId', str(siteId)))

      if not isinstance(urlMd5, basestring) or urlMd5 == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('urlMd5', str(urlMd5)))

      if contentType is None:
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('contentType', str(contentType)))

      dataDir = (rawDataDir if rawDataDir[-1] == os.path.sep else rawDataDir + os.path.sep) + siteId + os.path.sep + PathMaker(urlMd5).getDir()
      fileSuffix = CrawlerDataStorage.getFileSuffix(contentType)
      oldFileName = dataDir + os.path.sep + urlMd5 + fileSuffix
      if os.path.isfile(oldFileName):
        timePostfix = datetime.datetime.now().strftime(CrawlerDataStorage.DATETIME_SUFFIX_FORMAT)
        newFileName = dataDir + os.path.sep + urlMd5 + timePostfix + fileSuffix
        os.rename(oldFileName, newFileName)

        if log is not None:
          log.debug("File %s was renamed to %s", str(oldFileName), str(newFileName))

    except Exception, err:
      if log is not None:
        log.error(str(err))

#   # # Rename cookies resource file
#   #
#   # @param siteId - Site/Project ID
#   # @param urlMd5 - urlMd5 string
#   # @return - None
#   def renameCookiesFile(self, siteId, urlMd5):
#     if self.keepOldResource and isinstance(self.rawDataDir, basestring) and self.rawDataDir != "":
#       self.renameResourceFile(self.rawDataDir, siteId, urlMd5, dc_event.Content.CONTENT_COOKIES_CONTENT, self.logger)


  # # Rename all resource files
  #
  # @param urlPutList - list of a URLPut instances
  # @param rawDataDir - raw data directory
  # @param keepOldResource - boolean flag keep old resources
  # @param log - logger instance
  # @return - None
  @staticmethod
  def renameAllDataFiles(urlPutList, rawDataDir, keepOldResource, log=None):
    try:
      if not isinstance(urlPutList, list):
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('urlPutList', str(urlPutList)))

      if keepOldResource and isinstance(rawDataDir, basestring) or rawDataDir == "":
        for urlPutObj in urlPutList:
          CrawlerDataStorage.renameResourceFile(rawDataDir, urlPutObj.siteId, urlPutObj.urlMd5, urlPutObj.contentType, log)

    except Exception, err:
      if log is not None:
        log.error(str(err))


  # # Add URL put data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param contentType - type of the content
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addData(siteId, urlMd5, contentType, data, urlPutList, log=None):
    try:
      if not isinstance(siteId, basestring) or siteId == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('siteId', str(siteId)))

      if not isinstance(urlMd5, basestring) or urlMd5 == "":
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('urlMd5', str(urlMd5)))

      if contentType is None:
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('contentType', str(contentType)))

      if not isinstance(urlPutList, list):
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('urlPutList', str(urlPutList)))

      putDict = {CrawlerDataStorage.CONTENT_DATA_NAME: base64.b64encode(data)}
      urlPutList.append(dc_event.URLPut(siteId, urlMd5, contentType, putDict))

    except Exception, err:
      if log is not None:
        log.error(str(err))


  # # Add header content data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addHeaderContentData(siteId, urlMd5, data, urlPutList, log=None):
    CrawlerDataStorage.addData(siteId, urlMd5, dc_event.Content.CONTENT_HEADERS_CONTENT, data, urlPutList, log)


  # # Add raw content data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  @staticmethod
  def addRawContentData(siteId, urlMd5, data, urlPutList, log=None):
    CrawlerDataStorage.addData(siteId, urlMd5, dc_event.Content.CONTENT_RAW_CONTENT, data, urlPutList, log)


  # # Add tidy content data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addTidyContentData(siteId, urlMd5, data, urlPutList, log=None):
    CrawlerDataStorage.addData(siteId, urlMd5, dc_event.Content.CONTENT_TIDY_CONTENT, data, urlPutList, log)


  # # Add dynamic content data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addDynamicContentData(siteId, urlMd5, data, urlPutList, log=None):
    CrawlerDataStorage.addData(siteId, urlMd5, dc_event.Content.CONTENT_DYNAMIC_CONTENT, data, urlPutList, log)


  # # Add request content data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addRequestContentData(siteId, urlMd5, data, urlPutList, log=None):
    CrawlerDataStorage.addData(siteId, urlMd5, dc_event.Content.CONTENT_REQUESTS_CONTENT, data, urlPutList, log)


  # # Add cookies content data
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def addCookiesContentData(siteId, urlMd5, data, urlPutList, log=None):
    CrawlerDataStorage.addData(siteId, urlMd5, dc_event.Content.CONTENT_COOKIES_CONTENT, data, urlPutList, log)


  ## Put contents data list
  #
  # @param dbWrapper - DBTasksWrapper instance
  # @param urlPutList - list of a URLPut instances
  # @param log - logger instance
  # @return - None
  @staticmethod
  def putContentsList(dbWrapper, urlPutList, log=None):
    try:
      if not isinstance(urlPutList, list):
        raise Exception(CrawlerDataStorage.ERROR_INPUT_DATA % ('urlPutList', str(urlPutList)))

      if isinstance(dbWrapper, DBTasksWrapper):
        dbWrapper.putURLContent(urlPutList)

    except Exception, err:
      if log is not None:
        log.error(str(err))


  # # Add header content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @return - None
  def addHeaderContent(self, siteId, urlMd5, data):
    data = HTTPHeaders.dictToHeader(collections.OrderedDict(sorted(HTTPHeaders.headerToDict(data).items())))
    self.addHeaderContentData(siteId, urlMd5, data, self.urlPutList, self.logger)


  # # Add raw content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @return - None
  def addRawContent(self, siteId, urlMd5, data):
    self.addRawContentData(siteId, urlMd5, data, self.urlPutList, self.logger)


  # # Add tidy content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  def addTidyContent(self, siteId, urlMd5, data):
    self.addTidyContentData(siteId, urlMd5, data, self.urlPutList, self.logger)


  # # Add dynamic content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @return - None
  def addDynamicContent(self, siteId, urlMd5, data):
    self.addDynamicContentData(siteId, urlMd5, data, self.urlPutList, self.logger)


  # # Add request content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @return - None
  def addRequestContent(self, siteId, urlMd5, data):
    self.addRequestContentData(siteId, urlMd5, data, self.urlPutList, self.logger)


  # # Add cookies content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - urlMd5 string
  # @param data - data value string
  # @return - None
  def addCookiesContent(self, siteId, urlMd5, data):
    self.addCookiesContentData(siteId, urlMd5, data, self.urlPutList, self.logger)


  # Put contents
  #
  # @param - None
  # @return - None
  def putContents(self):
    # rename all resources files if necessary
    self.renameAllDataFiles(self.urlPutList, self.rawDataDir, self.keepOldResource, self.logger)
    # put contents
    self.putContentsList(self.dbWrapper, self.urlPutList, self.logger)
    self.urlPutList = []
