# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
DetectModified Class content main functional for 'DETECT_MODIFIED' property.

@package: dc_crawler
@file DetectModified.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import re
import copy
import hashlib

import app.Utils as Utils
import app.Consts as APP_CONSTS
from app.DateTimeType import DateTimeType
import dc.EventObjects as dc_events
import dc_crawler.Constants as CONSTS
from dc_crawler.Fetcher import Response
from dc_crawler.ResourceProcess import ResourceProcess
from dc_crawler.Exceptions import NotModifiedException
from dc_crawler.DBTasksWrapper import DBTasksWrapper

logger = Utils.MPLogger().getLogger()

class DetectModified(object):
  # # Constants error messages used in class
  ERROR_MSG_FOUND_NOT_MODIFIED_RESOURCE = "Found not modified resource."
  ERROR_MSG_BAD_VALUE = "Wrong value '%s' for argument '%s'"

  # # Constants used in class
  MODE_DISABLED = 0
  MODE_ENABLED = 1

  DEFAULT_MODE_VALUE = MODE_DISABLED
  SUPPORT_MODES = [MODE_DISABLED, MODE_ENABLED]

  ALGORITHM_GET_IF_NONE_MATCH_AND_IF_MOFIFIED_SINCE = 0
  ALGORITHM_GET_IF_NONE_MATCH_COMPARE_ETAG = 1
  ALGORITHM_GET_IF_MOFIFIED_SINCE_COMPARE_CDATE = 2
  ALGORITHM_HEAD_COMPARE_CDATE = 3
  ALGORITHM_GET_COMPARE_RAW_CONTENT_HASH = 4
  ALGORITHM_GET_COMPARE_RAW_CONTENT_LEN = 5

  DEFAULT_ALGORITHM_VALUE = ALGORITHM_GET_IF_NONE_MATCH_AND_IF_MOFIFIED_SINCE
  SUPPORT_ALGORITHMS = [ALGORITHM_GET_IF_NONE_MATCH_AND_IF_MOFIFIED_SINCE, ALGORITHM_GET_IF_NONE_MATCH_COMPARE_ETAG,
                        ALGORITHM_GET_IF_MOFIFIED_SINCE_COMPARE_CDATE, ALGORITHM_HEAD_COMPARE_CDATE,
                        ALGORITHM_GET_COMPARE_RAW_CONTENT_HASH, ALGORITHM_GET_COMPARE_RAW_CONTENT_LEN]

  BEHAVIOR_UPDATE_HTTP_CODE = 0
  BEHAVIOR_UPDATE_STATUS_TO_CRAWLED = 1
  BEHAVIOR_UPDATE_STATUS_TO_PROCESSED = 2
  BEHAVIOR_UPDATE_UDATE = 3

  DEFAULT_BEHAVIOR_VALUE = BEHAVIOR_UPDATE_HTTP_CODE
  SUPPORT_BEHAVIORS = [BEHAVIOR_UPDATE_HTTP_CODE, BEHAVIOR_UPDATE_STATUS_TO_CRAWLED, BEHAVIOR_UPDATE_STATUS_TO_PROCESSED, BEHAVIOR_UPDATE_UDATE]

  HEADER_OPTION_IF_NONE_MATCH = 'If-None-Match'
  HEADER_OPTION_IF_MOFIFIED_SINCE = 'If-Modified-Since'
  HEADER_OPTION_ETAG = 'ETag'

  METHOD_NAME_GET = 'get'
  METHOD_NAME_HEAD = 'head'

  ARGUMENT_URL_OBJECT = 'urlObj'


  ## Internal class for properties
  class DetectModifiedProperty(object):
    # # Constants used in class
    OPTION_MODE_NAME = 'mode'
    OPTION_ALGORITHM_NAME = 'algorithm'
    OPTION_BEHAVIOR_NAME = 'behavior'
    OPTION_URL_MASK_NAME = 'url_mask'


    # # Initialization
    def __init__(self, mode=None, algorithm=None, behavior=None):
      self.mode = mode
      self.algorithm = algorithm
      self.behavior = behavior


  ## Initialization
  #
  # @param siteProperties - site properties dictionary
  # @param url - URL string
  def __init__(self, siteProperties, url):
    self.detectModifiedProperty = self.getDetectModifiedProperty(siteProperties, url)
    self.prevUrlObj = None


  # # Is initialized for usage  bydetect modified
  #
  # @param - None
  # @return True if initialized for usage or otherwise False
  def isInitialized(self):
    return self.detectModifiedProperty is not None


  # # get DetectModifiedProperty instance or None
  #
  # @param siteProperties - site properties dictionary
  # @param url - URL string
  # @return DetectModifiedProperty instance
  def getDetectModifiedProperty(self, siteProperties, url):
    # variable for result
    ret = None
    try:
      if isinstance(siteProperties, dict) and APP_CONSTS.DETECT_MODIFIED in siteProperties and siteProperties[APP_CONSTS.DETECT_MODIFIED] != "":
        # load property
        properties = Utils.jsonLoadsSafe(jsonString=siteProperties[APP_CONSTS.DETECT_MODIFIED], default=siteProperties[APP_CONSTS.DETECT_MODIFIED], log=logger)
        if isinstance(properties, dict):
          properties = [properties]

        if isinstance(properties, list):
          for elem in properties:
            logger.debug("elem: %s", str(elem))
            if isinstance(elem, dict):
              # check allowed URL pattern
              if self.DetectModifiedProperty.OPTION_URL_MASK_NAME in elem and elem[self.DetectModifiedProperty.OPTION_URL_MASK_NAME] and \
                isinstance(elem[self.DetectModifiedProperty.OPTION_URL_MASK_NAME], list) and isinstance(url, basestring):
                # search URLs mask
                for urlMask in elem[self.DetectModifiedProperty.OPTION_URL_MASK_NAME]:
                  if isinstance(urlMask, basestring) and urlMask != "":
                    if re.search(pattern=urlMask, string=url, flags=re.U + re.I) is None:
                      # skipped as not allowed
                      continue
                    else:
                      logger.debug("Pattern '%s' passed for url: %s", str(urlMask), str(url))
                      break

              mode = self.DEFAULT_MODE_VALUE
              # get and check mode value
              if self.DetectModifiedProperty.OPTION_MODE_NAME in elem and elem[self.DetectModifiedProperty.OPTION_MODE_NAME]:
                mode = int(elem[self.DetectModifiedProperty.OPTION_MODE_NAME])
                if mode not in self.SUPPORT_MODES:
                  logger.error("Input mode '%s' not supported.", str(mode))
                  continue

              algorithm = self.DEFAULT_ALGORITHM_VALUE
              # get and check mode value
              if self.DetectModifiedProperty.OPTION_ALGORITHM_NAME in elem and elem[self.DetectModifiedProperty.OPTION_ALGORITHM_NAME]:
                algorithm = int(elem[self.DetectModifiedProperty.OPTION_ALGORITHM_NAME])
                if algorithm not in self.SUPPORT_ALGORITHMS:
                  logger.error("Input algorithm '%s' not supported.", str(algorithm))
                  continue

              behavior = self.DEFAULT_BEHAVIOR_VALUE
              # get and check algorithm value
              if self.DetectModifiedProperty.OPTION_BEHAVIOR_NAME in elem and elem[self.DetectModifiedProperty.OPTION_BEHAVIOR_NAME]:
                behavior = int(elem[self.DetectModifiedProperty.OPTION_BEHAVIOR_NAME])
                if behavior not in self.SUPPORT_BEHAVIORS:
                  logger.error("Input behavior '%s' not supported.", str(behavior))
                  continue

              logger.debug("mode = %s, algorithm = %s, behavior= %s", str(mode), str(algorithm), str(behavior))
              if mode == self.MODE_ENABLED:
                ret = self.DetectModifiedProperty(mode, algorithm, behavior)
                break

    except Exception, err:
      logger.error(str(err))

    return ret


  # # Prepare headers
  #
  # @param headers - request headers dictionary
  # @return prepared headers dictionary
  @staticmethod
  def prepareHeaders(headers):
    if isinstance(headers, dict):
      deleteOptionsList = []
      for key in headers.keys():
        if key.lower() == DetectModified.HEADER_OPTION_IF_NONE_MATCH.lower() or key.lower() == DetectModified.HEADER_OPTION_IF_MOFIFIED_SINCE.lower():
          deleteOptionsList.append(key)

      for elem in deleteOptionsList:
        del headers[elem]

    return headers


  # # Generate eTag string
  #
  # @param eTags contains list of eTags
  # @return eTag string
  def generateETagsString(self, eTags):
    # variable for result
    ret = ""
    if isinstance(eTags, list):
      ret = '","'.join(eTags)
    elif isinstance(eTags, basestring):
      ret = eTags
    elif eTags is None:
      ret = '*'

    ret = '"' + ret + '"'

    return ret


  # # Prepare request
  #
  # @param urlObj - URL instance
  # @param headers - request headers dictionary
  # @return modified headers dictionary if necessary
  def prepareRequest(self, urlObj, headers):
    # variable for result
    ret = headers
    try:
      if self.isInitialized() and isinstance(urlObj, dc_events.URL) and isinstance(headers, dict):
        # save URL object
        self.prevUrlObj = copy.deepcopy(urlObj)

        logger.debug("urlObj.eTag: %s, urlObj.lastModified: %s", str(urlObj.eTag), str(urlObj.lastModified))
        logger.debug("Input headers: %s", str(headers))

        # # prepare headers
        headers = self.prepareHeaders(headers)
        # change headers if necessary
        if isinstance(self.detectModifiedProperty, self.DetectModifiedProperty) and self.detectModifiedProperty.mode == self.MODE_ENABLED:
          if self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_NONE_MATCH_AND_IF_MOFIFIED_SINCE or \
            self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_NONE_MATCH_COMPARE_ETAG:
            # necessary add 'eTag'
            if urlObj.eTag is not None and urlObj.eTag != "":
              headers[self.HEADER_OPTION_IF_NONE_MATCH] = self.generateETagsString(urlObj.eTag)

          if self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_NONE_MATCH_AND_IF_MOFIFIED_SINCE or \
            self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_MOFIFIED_SINCE_COMPARE_CDATE:
            # necessary add 'If-modified-since'
            if urlObj.lastModified is not None and urlObj.lastModified != "":
              headers[self.HEADER_OPTION_IF_MOFIFIED_SINCE] = urlObj.lastModified

    except Exception, err:
      logger.error(str(err))

    logger.debug("Output headers: %s", str(headers))
    return ret


  # # Get HTTP method
  #
  # @param method, fetch HTTP method
  # @return HTTP method for apply
  def getHTTPMethod(self, default=None):
    #variable for result
    ret = DetectModified.METHOD_NAME_GET if default is None else default

    if self.isInitialized() and isinstance(self.detectModifiedProperty, self.DetectModifiedProperty) and \
      self.detectModifiedProperty.mode == DetectModified.MODE_ENABLED and self.detectModifiedProperty.algorithm == self.ALGORITHM_HEAD_COMPARE_CDATE:
      ret = self.METHOD_NAME_HEAD

    return ret


  # # Check is need to repeat request
  #
  # @param - None
  # @return True if need to repeat or otherwise False
  def isNeedRepeatRequest(self):
    return self.getHTTPMethod() == self.METHOD_NAME_HEAD


  ## Check is modified according to used algorithm
  #
  # @param response - URL fetch response
  # @return True if modified or otherwise False
  def isNotModified(self, response):
    # variable for result
    ret = False
    try:
      if isinstance(response, Response) and self.isInitialized():
        if self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_NONE_MATCH_AND_IF_MOFIFIED_SINCE:
          logger.debug("status_code = %s", str(ret))
          if response.status_code == CONSTS.HTTP_CODE_304:
            ret = True

        elif self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_NONE_MATCH_COMPARE_ETAG:
          if isinstance(response.headers, dict):
            for key, value in response.headers.items():
              if key.lower() == self.HEADER_OPTION_ETAG.lower():
                logger.debug("new eTag = %s, previous eTag = %s", str(value), str(self.prevUrlObj.eTag))
                if value == self.prevUrlObj.eTag:
                  ret = True
                break

        elif self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_IF_MOFIFIED_SINCE_COMPARE_CDATE:
          lastModified = ResourceProcess.getLastModified(response)
          logger.debug("lastModified = %s, previous lastModified = %s", str(lastModified), str(self.prevUrlObj.lastModified))

          lastModifiedDatetime = DateTimeType.parse(lastModified)
          previousLastModifiedDatetime = DateTimeType.parse(self.prevUrlObj.lastModified)
          logger.debug("lastModified = %s, previous lastModified = %s", str(lastModifiedDatetime), str(previousLastModifiedDatetime))
          if lastModifiedDatetime <= previousLastModifiedDatetime:
            ret = True

        elif self.detectModifiedProperty.algorithm == self.ALGORITHM_HEAD_COMPARE_CDATE:
          lastModified = ResourceProcess.getLastModified(response)
          logger.debug("lastModified = %s, previous CDate = %s", str(lastModified), str(self.prevUrlObj.CDate))

          lastModifiedDatetime = DateTimeType.parse(lastModified)
          previousCDateDatetime = DateTimeType.parse(self.prevUrlObj.CDate)
          logger.debug("lastModified = %s, previous CDate = %s", str(lastModifiedDatetime), str(previousCDateDatetime))
          if lastModifiedDatetime <= previousCDateDatetime:
            ret = True

        elif self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_COMPARE_RAW_CONTENT_HASH and isinstance(response.rendered_unicode_content, basestring):
          rawContentMd5 = hashlib.md5(response.rendered_unicode_content).hexdigest()
          logger.debug("new rawContentMd5 = %s, previous rawContentMd5 = %s", str(rawContentMd5), str(self.prevUrlObj.rawContentMd5))
          if rawContentMd5 == self.prevUrlObj.rawContentMd5:
            ret = True

        elif self.detectModifiedProperty.algorithm == self.ALGORITHM_GET_COMPARE_RAW_CONTENT_LEN and isinstance(response.rendered_unicode_content, basestring):
          size = len(response.rendered_unicode_content)
          logger.debug("new size = %s, previous size = %s", str(size), str(self.prevUrlObj.size))
          if size == self.prevUrlObj.size:
            ret = True

    except Exception, err:
      logger.error(str(err))

    logger.debug("Is not modified = %s", str(ret))
    return ret


  ## Check modified
  #
  # @param detectModified - DetectModified instance
  # @param response - URL fetch response
  # @return - None
  @staticmethod
  def checkModified(detectModified, response):
    if isinstance(detectModified, DetectModified) and isinstance(response, Response) and detectModified.isInitialized() and \
      detectModified.detectModifiedProperty.mode == DetectModified.MODE_ENABLED and detectModified.isNotModified(response):
      raise NotModifiedException(message=DetectModified.ERROR_MSG_FOUND_NOT_MODIFIED_RESOURCE, detectModified=detectModified, response=response)


  # # Update response if necessary
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - hash MD5 for URL string
  # @param dbWrapper - DBWrapper instance
  # @param detectModified - DetectModified instance
  # @param response - URL fetch response
  # @return result tuple (httpCode, status, updateUDate)
  @staticmethod
  def extractResult(siteId, urlMd5, dbWrapper, detectModified, response):
    # variable for result
    httpCode = response.status_code if response.status_code > 0 else CONSTS.HTTP_CODE_400
    status = dc_events.URL.STATUS_CRAWLED
    updateUDate = True
    try:
      if isinstance(detectModified, DetectModified) and detectModified.isInitialized() and \
        detectModified.detectModifiedProperty.mode == DetectModified.MODE_ENABLED and isinstance(response, Response):
#         logger.info("Try extract result for detectModified:\n%s", Utils.varDump(detectModified, stringifyType=0))

        isNotModified = detectModified.isNotModified(response)
        # apply behavior
        if detectModified.detectModifiedProperty.behavior == DetectModified.BEHAVIOR_UPDATE_HTTP_CODE and not isNotModified and \
          response.status_code == CONSTS.HTTP_CODE_200:
          httpCode = CONSTS.HTTP_CODE_308

        elif detectModified.detectModifiedProperty.behavior == DetectModified.BEHAVIOR_UPDATE_STATUS_TO_CRAWLED and isNotModified:
          status = dc_events.URL.STATUS_CRAWLED

        elif detectModified.detectModifiedProperty.behavior == DetectModified.BEHAVIOR_UPDATE_STATUS_TO_PROCESSED and isNotModified:
          status = dc_events.URL.STATUS_CRAWLED
          if DetectModified.isExistsContent(siteId, urlMd5, dbWrapper):
            status = dc_events.URL.STATUS_PROCESSED

        elif detectModified.detectModifiedProperty.behavior == DetectModified.BEHAVIOR_UPDATE_UDATE and isNotModified:
          updateUDate = False

    except Exception, err:
      logger.error(str(err))

    logger.debug("Return httpCode = %s, status = %s, updateUDate = %s", str(httpCode), str(status), str(updateUDate))
    return httpCode, status, updateUDate


  # # Check is exists content
  #
  # @param siteId - Site/Project ID
  # @param urlMd5 - hash MD5 for URL string
  # @param dbWrapper - DBWrapper instance
  # @return True if exists or otherwise False
  @staticmethod
  def isExistsContent(siteId, urlMd5, dbWrapper):
    #variable for result
    ret = False
    try:
      if not isinstance(siteId, basestring) or siteId == "":
        raise DetectModified.ERROR_MSG_BAD_VALUE % (str(siteId), 'siteId')

      if not isinstance(urlMd5, basestring) or urlMd5 == "":
        raise DetectModified.ERROR_MSG_BAD_VALUE % (str(urlMd5), 'urlMd5')

      if not isinstance(dbWrapper, DBTasksWrapper):
        raise DetectModified.ERROR_MSG_BAD_VALUE % (str(dbWrapper), 'dbWrapper')

      urlContentRequest = dc_events.URLContentRequest(siteId,
                                                      urlMd5,
                                                      dc_events.URLContentRequest.CONTENT_TYPE_PROCESSED,
                                                      dc_events.URLContentRequest.URL_TYPE_MD5)


      urlContentResponse = dbWrapper.urlContent([urlContentRequest])
      logger.info("!!! urlContentResponse:\n%s", Utils.varDump(urlContentResponse, stringifyType=0))

      if isinstance(urlContentResponse, list) and len(urlContentResponse) > 0 and \
        isinstance(urlContentResponse[0].processedContents, list) and len(urlContentResponse[0].processedContents) > 0:
        ret = True

    except Exception, err:
      logger.error(str(err))

    return ret
