# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
HTTPRedirectResolver Class content main functional for resolve redirects

@package: dc_crawler
@file HTTPRedirectResolver.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2017 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import re

import app.Consts as APP_CONSTS
import app.Utils as Utils
from app.Utils import varDump
from app.Utils import getTracebackInfo
from app.ContentEvaluator import ContentEvaluator
from dc_crawler.Fetcher import BaseFetcher
import dc_crawler.Constants as CONSTS
from dc_crawler.HTTPCookieResolver import HTTPCookieResolver
from dc_crawler.HTTPProxyResolver import HTTPProxyResolver
from dc_crawler.RequestsRedirectWrapper import RequestsRedirectWrapper
from dc_crawler.Exceptions import BadContentTypeException
from dc_crawler.DetectModified import DetectModified

logger = Utils.MPLogger().getLogger()

class HTTPRedirectResolver(object):
  # # Constants
  DEFAULT_VALUE_COOKIES_STAGE = HTTPCookieResolver.STAGE_DEFAULT

  # #Constants error message
  ERROR_INITIALIZATION = "Initialization class '%s' was failed. Error: %s"
  ERROR_BAD_TYPE_PROPERTY_VALUE = "Wrong type (%s) of property value: %s"
  ERROR_BAD_TYPE_HEADERS_VALUE = "Wrong type (%s) of headers: %s"
  ERROR_BAD_PROPERTY_VALUE = "Not support value '%s' for property '%s'"
  ERROR_BAD_STATUS_CODE_VALUE = "Not allowed status code '%s'. Allowed list: %s"


  # # Internal class for redirect property
  class RedirectProperty(object):
    # #Constants used in class
    PROPERTY_METHOD_NAME = "METHOD"
    PROPERTY_URL_PATTERN_LIST = "URL"
    PROPERTY_MAX_REDIRECTS = "MAX"
    PROPERTY_TYPES_LIST = "TYPES"
    PROPERTY_INSERT_AS_NEW = "insert_as_new"
    PROPERTY_UPDATE_OLD = "update_old"
    PROPERTY_STAGE = "Stage"

    METHOD_NAME_HEAD = 'HEAD'
    METHOD_NAME_SGET = 'SGET'
    METHOD_NAME_DGET = 'DGET'

    METHOD_NAME_GET = 'GET'

    VALUE_INSERT_AS_NEW_DO_NOT_INSERT = 0
    VALUE_INSERT_AS_NEW_INSERT = 1

    VALUE_STAGE_OFF = 0
    VALUE_STAGE_CRAWLER = 1
    VALUE_STAGE_COLLECT_URL = 2

    DEFAULT_VALUE_METHOD_NAME = METHOD_NAME_HEAD
    DEFAULT_VALUE_URL_PATTERN_LIST = ['.*']
    DEFAULT_VALUE_MAX_REDIRECTS = 10
    DEFAULT_VALUE_TYPES_LIST = [301, 302, 303, 304]
    DEFAULT_VALUE_INSERT_AS_NEW = VALUE_INSERT_AS_NEW_DO_NOT_INSERT
    DEFAULT_VALUE_UPDATE_OLD = {}
    DEFAULT_VALUE_STAGE = VALUE_STAGE_CRAWLER

    SUPPORT_METHOD_NAMES = [METHOD_NAME_HEAD, METHOD_NAME_SGET, METHOD_NAME_DGET]
    SUPPORT_STAGE_VALUES = [VALUE_STAGE_OFF, VALUE_STAGE_CRAWLER, VALUE_STAGE_COLLECT_URL]


    def __init__(self, methodName=None, urlPatternList=None, maxRedirects=None, typesList=None, insertAsNew=None, updateOld=None, stage=None):
      self.methodName = self.DEFAULT_VALUE_METHOD_NAME if methodName is None else methodName
      self.urlPatternList = self.DEFAULT_VALUE_URL_PATTERN_LIST if urlPatternList is None else urlPatternList
      self.maxRedirects = self.DEFAULT_VALUE_MAX_REDIRECTS if maxRedirects is None else maxRedirects
      self.typesList = self.DEFAULT_VALUE_TYPES_LIST if typesList is None else typesList
      self.insertAsNew = self.DEFAULT_VALUE_INSERT_AS_NEW if insertAsNew is None else insertAsNew
      self.updateOld = self.DEFAULT_VALUE_UPDATE_OLD if updateOld is None else updateOld
      self.stage = self.DEFAULT_VALUE_STAGE if stage is None else stage


  # # Internal class for insensitive dictionary
  class CaseInsensitivePropertyDict(object):


    # Initialization
    def __init__(self, dataDict):
      self.dataDict = {key.lower():value for key, value in dataDict.items()}


    def __contains__(self, key):
      return key.lower() in self.dataDict


    def __delitem__(self, key):
      del self.dataDict[key.lower()]


    def __getitem__(self, key):
      return self.dataDict[key.lower()]


    def __setitem__(self, key, value):
      self.dataDict[key.lower()] = value


  # Initialization
  # @param siteProperties - site properties dictionary
  # @param fetchType - fetcher type
  # @param dbWrapper - DBTaskWrapper instance
  # @param connectionTimeout - connection timeout
  # @param siteId- site ID
  # @param stage - crawler stage for apply
  # @param cookiesStage - cookies stage
  def __init__(self, siteProperties=None, fetchType=BaseFetcher.TYP_NORMAL, dbWrapper=None, siteId=None,
               connectionTimeout=CONSTS.CONNECTION_TIMEOUT, stage=None, cookiesStage=None):
    self.redirectProperty = self.__loadProperty(siteProperties)
    self.siteProperties = siteProperties
    self.fetchType = fetchType
    self.dbWrapper = dbWrapper
    self.siteId = siteId
    self.connectionTimeout = connectionTimeout
    self.stage = self.RedirectProperty.DEFAULT_VALUE_STAGE if stage is None else stage
    self.cookiesStage = self.DEFAULT_VALUE_COOKIES_STAGE if cookiesStage is None else cookiesStage


  # # get redirect property
  #
  # @param propertyDict - contains dictionary
  # @return RedirectProperty instance
  def getRedirectProperty(self, propertyDict):
  # variable for result
    redirectProperty = HTTPRedirectResolver.RedirectProperty()
    try:
      if not isinstance(propertyDict, dict):
        raise Exception(self.ERROR_BAD_TYPE_PROPERTY_VALUE % (str(type(propertyDict)), varDump(propertyDict)))

      caseInsensitivePropertyDict = self.CaseInsensitivePropertyDict(propertyDict)

      # extract method name
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_METHOD_NAME in caseInsensitivePropertyDict:
        if caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_METHOD_NAME] in \
          HTTPRedirectResolver.RedirectProperty.SUPPORT_METHOD_NAMES:
          redirectProperty.methodName = caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_METHOD_NAME]
        else:
          raise Exception(self.ERROR_BAD_PROPERTY_VALUE % \
                          (varDump(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_METHOD_NAME]),
                           str(HTTPRedirectResolver.RedirectProperty.PROPERTY_METHOD_NAME)))

      # extract url pattern list
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_URL_PATTERN_LIST in caseInsensitivePropertyDict:
        if isinstance(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_URL_PATTERN_LIST], list):
          redirectProperty.urlPatternList = \
            caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_URL_PATTERN_LIST]

          for i in xrange(len(redirectProperty.urlPatternList)):
            if isinstance(redirectProperty.urlPatternList[i], dict):
              redirectProperty.urlPatternList[i] = self.getRedirectProperty(redirectProperty.urlPatternList[i])

        else:
          raise Exception(self.ERROR_BAD_PROPERTY_VALUE % \
                          (varDump(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_URL_PATTERN_LIST]),
                           str(HTTPRedirectResolver.RedirectProperty.PROPERTY_URL_PATTERN_LIST)))

      # extract max redirects value
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_MAX_REDIRECTS in caseInsensitivePropertyDict:
        redirectProperty.maxRedirects = \
          int(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_MAX_REDIRECTS])

      # extract redirect types list
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_TYPES_LIST in caseInsensitivePropertyDict:
        if isinstance(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_TYPES_LIST], list):
          redirectProperty.typesList = caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_TYPES_LIST]
        else:
          raise Exception(self.ERROR_BAD_PROPERTY_VALUE % \
                          (varDump(propertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_TYPES_LIST]),
                           str(HTTPRedirectResolver.RedirectProperty.PROPERTY_TYPES_LIST)))

      # extract insert as new option
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_INSERT_AS_NEW in caseInsensitivePropertyDict:
        redirectProperty.insertAsNew = int(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_INSERT_AS_NEW])

      # extract update old option
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_UPDATE_OLD in caseInsensitivePropertyDict:
        if isinstance(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_UPDATE_OLD], dict):
          redirectProperty.updateOld = caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_UPDATE_OLD]

      # extract stage option
      if HTTPRedirectResolver.RedirectProperty.PROPERTY_STAGE in caseInsensitivePropertyDict:
        if isinstance(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_STAGE], dict):
          redirectProperty.stage = int(caseInsensitivePropertyDict[HTTPRedirectResolver.RedirectProperty.PROPERTY_STAGE])

    except Exception, err:
      logger.error(self.ERROR_INITIALIZATION, self.__class__.__name__, str(err))
      logger.info(getTracebackInfo())

    return redirectProperty


  # # load property from input json
  #
  # @param siteProperties - site properties dictionary
  # @return RedirectProperty instance
  def __loadProperty(self, siteProperties):
    # variable for result
    redirectProperty = HTTPRedirectResolver.RedirectProperty()

    propertyString = None
    if APP_CONSTS.HTTP_REDIRECT_RESOLVER in siteProperties and siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER] != "":
      propertyString = siteProperties[APP_CONSTS.HTTP_REDIRECT_RESOLVER]
      logger.debug("Found property: '%s': %s", str(APP_CONSTS.HTTP_REDIRECT_RESOLVER), str(propertyString))

      redirectProperty = self.getRedirectProperty(Utils.jsonLoadsSafe(jsonString=propertyString, default=propertyString, log=logger))

    return redirectProperty


  # # check is allowed url use incming pattern list
  #
  # @param url - url string
  # @param patterns - regular expression pattern list
  # @return True if allowed or False - otherwise
  @staticmethod
  def isAllowedUrl(url, patterns):
    # variable for result
    ret = True
    try:
#       logger.debug("patterns: '%s'", str(patterns))
      if isinstance(patterns, list):
        for pattern in patterns:
          if isinstance(pattern, basestring):
            ret = False
            if re.search(pattern, url, re.UNICODE + re.IGNORECASE) is not None:
              logger.debug("pattern: '%s' allowed for '%s'", str(pattern), str(url))
              ret = True
              break

    except Exception, err:
      logger.error(str(err))

    return ret


  # # resolve redirect URL
  #
  # @param url - the URL to fetch
  # @param headers - request headers dictionary
  # @param method - fetch by HTTP method
  # @param timeout - request timeout(seconds)
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxies - proxies settings dictionary
  # @param auth - basic auth setting, tuple of name and password
  # @param postData - post data, used only when method is post
  # @param allowedContentTypes - allowed_content_types list
  # @param maxResourceSize - max resource size
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dictionary
  # @param depth - depth value
  # @param macro - macro code
  # @return resolved result url
  def resolveRedirectUrl(self, url, headers, timeout=None, allowRedirects=True, proxies=None, auth=None, postData=None, allowedContentTypes=None,
                         maxResourceSize=None, maxRedirects=RedirectProperty.DEFAULT_VALUE_MAX_REDIRECTS, filters=None, depth=None, macro=None):

    # variable for result
    ret = None

    logger.debug("Input url: %s \nheaders: %s", str(url), varDump(headers))
    try:
      res = self.fetchUrl(url=url, headers=headers, timeout=timeout, allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData,
                          allowedContentTypes=allowedContentTypes, maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters, depth=depth,
                          macro=macro)
      if res is not None:
        ret = res.url

    except Exception, err:
      logger.debug("Resolve redirect for URL '%s' failed by reason: %s", str(url), str(err))
      logger.info(Utils.getTracebackInfo())

    return ret


  # # fetch URL
  #
  # @param url - the URL to fetch
  # @param headers - request headers dictionary
  # @param method - fetch by HTTP method
  # @param timeout - request timeout(seconds)
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxies - proxies settings dictionary
  # @param auth - basic auth setting, tuple of name and password
  # @param postData - post data, used only when method is post
  # @param allowedContentTypes - allowed_content_types list
  # @param maxResourceSize - max resource size
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dictionary
  # @param depth - depth value
  # @param macro - macro code
  # @param fetchAlways - boolean flag fetch always
  # @param useDetectModified - boolean flag use detect modified
  # @return URL fetch response
  def fetchUrl(self, url, headers, method=RedirectProperty.METHOD_NAME_GET, timeout=None, allowRedirects=True, proxies=None, auth=None, postData=None,
               allowedContentTypes=None, maxResourceSize=None, maxRedirects=RedirectProperty.DEFAULT_VALUE_MAX_REDIRECTS, filters=None, depth=None, macro=None,
               fetchAlways=True, useDetectModified=False, **kwarg):
    # variable for result
    ret = None
    logger.debug("Input url: %s \nheaders: %s", str(url), varDump(headers))
    logger.debug("fetchType = %s, fetchAlways = %s, useDetectModified = %s", str(self.fetchType), str(fetchAlways), str(useDetectModified))

    # get proxy settings
    proxies = HTTPProxyResolver.getProxySettings(siteProperties=self.siteProperties, siteId=self.siteId, url=url, dbWrapper=self.dbWrapper, log=logger,
                                                 default=proxies)
    logger.debug("Use proxies: %s", varDump(proxies, stringifyType=0))

    # Create instance detect modified
    detectModified = DetectModified(self.siteProperties, url)
    # Try usage detect modified
    if useDetectModified and detectModified.isInitialized() and DetectModified.ARGUMENT_URL_OBJECT in kwarg:
      # change headers
      headers = detectModified.prepareRequest(kwarg[DetectModified.ARGUMENT_URL_OBJECT], headers)
      # get HTTP method
      method = detectModified.getHTTPMethod()
      # make fetch URL
      ret = self.__fetch(url=url, headers=headers, method=method, timeout=timeout, allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData,
                         allowedContentTypes=allowedContentTypes, maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters,
                         depth=depth, macro=macro, fetchType=self.fetchType)

      # check modified if not modified throw exception NotModifiedException
      DetectModified.checkModified(detectModified, ret)

      needRepeat = detectModified.isNeedRepeatRequest()
      logger.debug("needRepeat = %s", str(needRepeat))

      if needRepeat:
        # # prepare headers
        headers = DetectModified.prepareHeaders(headers)
        # make fetch URL
        ret = self.__fetch(url=url, headers=headers, method=DetectModified.METHOD_NAME_GET, timeout=timeout, allowRedirects=allowRedirects, proxies=proxies,
                           auth=auth, postData=postData, allowedContentTypes=allowedContentTypes, maxResourceSize=maxResourceSize, maxRedirects=maxRedirects,
                           filters=filters, depth=depth, macro=macro, fetchType=self.fetchType)

    else:
      # Try usage redirect resolver algorithms
      redirectProperty = self.getRedirectPropertyObject(url)
      if isinstance(redirectProperty, HTTPRedirectResolver.RedirectProperty):
        logger.debug("Found property for stage = %s, ", str(redirectProperty.stage))

        if self.stage == redirectProperty.stage:
          logger.debug("It's allowed stage... Start resolving...")
          if redirectProperty.stage == self.RedirectProperty.VALUE_STAGE_CRAWLER and \
            redirectProperty.methodName == HTTPRedirectResolver.RedirectProperty.METHOD_NAME_HEAD:
            if self.fetchType == BaseFetcher.TYP_NORMAL:
              redirectProperty.methodName = HTTPRedirectResolver.RedirectProperty.METHOD_NAME_SGET
            elif self.fetchType == BaseFetcher.TYP_DYNAMIC:
              redirectProperty.methodName = HTTPRedirectResolver.RedirectProperty.METHOD_NAME_DGET

          ret = self.__resolveRedirect(url=url, headers=headers, timeout=timeout, allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData,
                                       allowedContentTypes=allowedContentTypes, maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters,
                                       depth=depth, macro=macro, redirectProperty=redirectProperty)
        else:
          logger.debug("It's not allowed stage (%s). Resolving skipped...", str(self.stage))

      # default fetch URL
      if ret is None and fetchAlways:
        logger.debug("Got flag fetch always.")
        # get proxy settings
        proxies = HTTPProxyResolver.getProxySettings(siteProperties=self.siteProperties, siteId=self.siteId, url=url, dbWrapper=self.dbWrapper, log=logger,
                                                     default=proxies)
        # make fetch
        ret = self.__fetch(url=url, headers=headers, method=method, timeout=timeout, allowRedirects=allowRedirects, proxies=proxies, auth=auth,
                           postData=postData, allowedContentTypes=allowedContentTypes, maxResourceSize=maxResourceSize, maxRedirects=maxRedirects,
                           filters=filters, depth=depth, macro=macro, fetchType=self.fetchType)

    # Execution 'REPLACE' property if necessary
    if ret is not None:
      ret = ContentEvaluator.execute(siteProperties=self.siteProperties, dbWrapper=self.dbWrapper, siteId=self.siteId, url=ret.url, response=ret)

    # Below usage check for both cases
    # Check is allowed content-type
    if ret is not None and ret.status_code == CONSTS.HTTP_CODE_200:
      contentType = self.getContentType(ret)
      logger.debug("contentType = %s", str(contentType))
      logger.debug("ret.content_type = %s", str(ret.content_type))
      if isinstance(allowedContentTypes, list) and isinstance(contentType, basestring) and contentType != "" and contentType not in allowedContentTypes:
        raise BadContentTypeException(RequestsRedirectWrapper.ERROR_BAD_CONTENT_TYPE_VALUE % (str(contentType), str(allowedContentTypes)), response=ret)

    return ret


  # # get content type from response object after fetch
  #
  # @param response - response object after fetch
  # @return content-type string
  @staticmethod
  def getContentType(response):
    # variable for result
    ret = ''
    try:
      logger.debug("!!! getContentType() response: %s, type = %s", str(response), str(type(response)))

      if response is not None and response.headers is not None:
        logger.debug("!!! response.headers: %s, type = %s", str(response.headers), str(type(response.headers)))

        if isinstance(response.headers, basestring):
          for line in response.headers.splitlines():
            if RequestsRedirectWrapper.CONTENT_TYPE_HEADER_NAME.lower() in line.lower():
              ret = line.lower().split(";")[0]
        else:
          ret = RequestsRedirectWrapper.getContentType(response)

        logger.debug("!!! getContentType() ret: '%s'", str(ret))

    except Exception, err:
      logger.error(str(err))

    return ret


  # # resolve redirect
  #
  # @param url - the URL to fetch
  # @param headers - request headers dictionary
  # @param timeout - request timeout(seconds)
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxies - proxies settings dictionary
  # @param auth - basic auth setting, tuple of name and password
  # @param postData - post data, used only when method is post
  # @param allowedContentTypes - allowed_content_types list
  # @param maxResourceSize - max resource size
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dictionary
  # @param depth - depth value
  # @param macro - macro code
  # @param redirectProperty - RedirectProperty instance
  # @return resolved result URL
  def __resolveRedirect(self, url, headers, timeout=None, allowRedirects=True, proxies=None, auth=None, postData=None, allowedContentTypes=None,
                        maxResourceSize=None, maxRedirects=RedirectProperty.DEFAULT_VALUE_MAX_REDIRECTS, filters=None, depth=None, macro=None,
                        redirectProperty=None):
    # variable for result
    ret = None
    logger.debug("type(redirectProperty) = %s", str(type(redirectProperty)))
    if isinstance(redirectProperty, HTTPRedirectResolver.RedirectProperty):
      logger.debug("type is GOOD!!!")

      if redirectProperty.methodName == HTTPRedirectResolver.RedirectProperty.METHOD_NAME_HEAD:
        # method 'HEAD' execution
        ret = self.__fetch(url=url, method=HTTPRedirectResolver.RedirectProperty.METHOD_NAME_HEAD, headers=headers, timeout=timeout,
                           allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData, allowedContentTypes=allowedContentTypes,
                           maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters, depth=depth, macro=macro, fetchType=self.fetchType)

      elif redirectProperty.methodName == HTTPRedirectResolver.RedirectProperty.METHOD_NAME_SGET:
        # method 'GET' for static fetcher type execution
        ret = self.__fetch(url=url, method=HTTPRedirectResolver.RedirectProperty.METHOD_NAME_GET, headers=headers, timeout=timeout,
                           allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData, allowedContentTypes=allowedContentTypes,
                           maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters, depth=depth, macro=macro,
                           fetchType=BaseFetcher.TYP_NORMAL)

      elif redirectProperty.methodName == HTTPRedirectResolver.RedirectProperty.METHOD_NAME_DGET:
        # method 'GET' for dynamic fetcher type execution
        ret = self.__fetch(url=url, method=HTTPRedirectResolver.RedirectProperty.METHOD_NAME_GET, headers=headers, timeout=timeout,
                           allowRedirects=allowRedirects, proxies=proxies, auth=auth, postData=postData, allowedContentTypes=allowedContentTypes,
                           maxResourceSize=maxResourceSize, maxRedirects=maxRedirects, filters=filters, depth=depth, macro=macro,
                           fetchType=BaseFetcher.TYP_DYNAMIC)

    return ret


  # # make fetch
  #
  # @param url - the url to fetch
  # @param headers - request headers dict
  # @param method - fetch by HTTP method
  # @param timeout - request timeout(seconds)
  # @param allowRedirects - boolean flag allowed redirects
  # @param proxies - proxies settings dictionary
  # @param auth - basic auth setting, tuple of name and password
  # @param postData - post data, used only when method is post
  # @param allowedContentTypes - allowed_content_types list
  # @param maxResourceSize - max resource size
  # @param maxRedirects - max allowed redirects count
  # @param filters - filters dict
  # @param depth - depth value
  # @param macro - macro code
  # @param fetchType - fetch type
  # @return result URL
  def __fetch(self, url, headers, method=RedirectProperty.METHOD_NAME_GET, timeout=None, allowRedirects=True, proxies=None, auth=None, postData=None,
              allowedContentTypes=None, maxResourceSize=None, maxRedirects=RedirectProperty.DEFAULT_VALUE_MAX_REDIRECTS, filters=None,
              depth=None, macro=None, fetchType=BaseFetcher.TYP_NORMAL):

    requestsRedirectWrapper = RequestsRedirectWrapper(dbWrapper=self.dbWrapper, siteId=self.siteId, siteProperties=self.siteProperties)
    # get fetcher instance
    fetcher = BaseFetcher.get_fetcher(fetchType, None, requestsRedirectWrapper)
    fetcher.connectionTimeout = self.connectionTimeout
    # fetch execution
    return fetcher.open(url=url, method=method, headers=headers, timeout=timeout, allow_redirects=allowRedirects, proxies=proxies, auth=auth, data=postData,
                        log=logger, allowed_content_types=allowedContentTypes, max_resource_size=maxResourceSize, max_redirects=maxRedirects, filters=filters,
                        depth=depth, macro=macro)


  # # get redirect property
  #
  # @param url - the URL string
  # @return redirect property
  def getRedirectPropertyObject(self, url):
    # variable for result
    ret = None
    try:
      if isinstance(self.redirectProperty, HTTPRedirectResolver.RedirectProperty):
        for urlPatternElem in self.redirectProperty.urlPatternList:
          if isinstance(urlPatternElem, HTTPRedirectResolver.RedirectProperty) and HTTPRedirectResolver.isAllowedUrl(url, urlPatternElem.urlPatternList):
            ret = urlPatternElem
            break

        if ret is None and HTTPRedirectResolver.isAllowedUrl(url, self.redirectProperty.urlPatternList):
          ret = self.redirectProperty

    except Exception , err:
      logger.error(str(err))

    return ret


  # # check flag insert as new
  #
  # @param url - the URL string
  # @return boolean flag insert as new value
  def isInsertAsNewUrl(self, url):
    # variable for result
    ret = False
    try:
      redirectProperty = self.getRedirectPropertyObject(url)
      if isinstance(redirectProperty, HTTPRedirectResolver.RedirectProperty):
        logger.debug("redirectProperty.insertAsNew = %s", str(redirectProperty.insertAsNew))
        ret = bool(redirectProperty.insertAsNew)

    except Exception , err:
      logger.error(str(err))

    return ret


  # # get update old values
  #
  # @param url - the URL string
  # @return update old values dictionary from property dictionary
  def getUpdateOldValues(self, url):
    # variable for result
    ret = {}
    try:
      redirectProperty = self.getRedirectPropertyObject(url)
      if isinstance(redirectProperty, HTTPRedirectResolver.RedirectProperty):
        logger.debug("redirectProperty.updateOld = %s", str(redirectProperty.updateOld))
        if isinstance(redirectProperty.updateOld, dict) and len(redirectProperty.updateOld) > 0:
          ret = redirectProperty.updateOld

    except Exception , err:
      logger.error(str(err))

    return ret


  # # update URL object
  #
  # @param urlObj - URLs objects instance
  # @param url - the URL string
  # @return - None
  def updateUrl(self, urlObj, url):
    if urlObj is not None:
      valuesDict = self.getUpdateOldValues(url)
      for key, value in valuesDict.items():
        for attrName in urlObj.__dict__.keys():
          if attrName.lower() == key.lower():
            setattr(urlObj, attrName, value)


  # # get stage value
  #
  # @param url - the URL string
  # @return numeric stage value
  def getStageValue(self, url):
    # variable for result
    ret = self.RedirectProperty.DEFAULT_VALUE_STAGE
    try:
      redirectProperty = self.getRedirectPropertyObject(url)
      if isinstance(redirectProperty, HTTPRedirectResolver.RedirectProperty):
        logger.debug("redirectProperty.stage = %s", str(redirectProperty.stage))
        if int(redirectProperty.stage) in self.RedirectProperty.SUPPORT_STAGE_VALUES:
          ret = int(redirectProperty.stage)

    except Exception , err:
      logger.error(str(err))

    return ret
