# coding: utf-8
"""
HCE project,  Python bindings, Distributed Tasks Manager application.
MimeTypeExtractor Class content main handler for 'MIME_TYPE_AUTO_DETECT' property.

@package: dc_crawler
@file MimeTypeExtractor.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2018 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import re
import magic
import requests

import dc.EventObjects as dc_event
from dc_crawler.CrawledResource import CrawledResource
import app.Consts as APP_CONSTS
import app.Utils as Utils


class MimeTypeExtractor(object):
  # # Constants
  OPTION_URL_EXPRESSION = "url_expression"
  OPTION_MODE = "mode"
  OPTION_URL_TYPES = "url_types"
  OPTION_URL_PARENT = "url_parent"
  OPTION_CONTENT_TYPES = "content_types"

  URL_PARENT_TYPE_VALUE_ONLY_PARENT = 0
  URL_PARENT_TYPE_VALUE_ONLY_NOT_PARENT = 1
  
  MODE_VALUE_ALWAYS = 0
  MODE_VALUE_NOT_DEFINED_FROM_HTTP_HEADER_RESPONSE = 1
  MODE_VALUE_MATCH_CONTENT_TYPES = 2
  MODE_VALUE_NOT_MATCH_CONTENT_TYPES =3
  
  DEFAULT_VALUE_MODE = MODE_VALUE_ALWAYS

  DEFAULT_VALUE_HEAD_TIMEOUT = 1
  HEADERS_CONTENT_TYPE_OPTIONS_NAME = 'content-type'

  # # Constants error messages
  ERROR_INPUT_TYPE = "Wrong input type %s instead expected %s"

  # # Constants messages
  RESULT_CHECK_MESSAGE = "Check '%s': %s, status: '%s', url: %s"
  RESULT_SUCCESS = "SUCCESS"
  RESULT_FAIL = "FAIL"


  # # Initialization
  def __init__(self):
    pass


  # # Detect mime
  #
  # @param crawledResource - crawler resource instance
  # @param contentTypeMap - content type map
  # @param log - logger instance
  # @returns detected mime or None
  @staticmethod
  def detect(crawledResource, contentTypeMap=None, log=None):
    # variable for result
    ret = None
    try:
      if not isinstance(crawledResource, CrawledResource):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(crawledResource)), str(type(CrawledResource()))))

      if crawledResource.dynamic_fetcher_type:
        rawUnicodeContent = crawledResource.meta_content
      else:
        rawUnicodeContent = crawledResource.binary_content

      if rawUnicodeContent is not None:
        ret = magic.from_buffer(str(rawUnicodeContent), mime=True)
        if log is not None:
          log.debug("Detected: %s", str(ret))

      if isinstance(contentTypeMap, dict) and ret in contentTypeMap:
        if log is not None:
          log.debug("Mime type replaced from '%s' to '%s'", str(ret), str(contentTypeMap[ret]))

        ret = str(contentTypeMap[ret])

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  # # Check is allowed the one rule element
  #
  # @param crawledResource - crawler resource instance
  # @param propertyDict - property dictionary
  # @param urlObj - URLs object instance
  # @param log - logger instance
  # @return True if allowed or otherwise False
  @staticmethod
  def isAllowedRuleElement(crawledResource, propertyDict, urlObj, log=None):
    # Internal function
    #
    # @param name - name of the option
    # @param value - value of the option
    # @param url - url string
    # @param log - logger instance
    # @param passed - boolean flag
    # @return - None
    def printStatus(name, value, url, log, passed=True):
      if log is not None:
        log.debug(MimeTypeExtractor.RESULT_CHECK_MESSAGE, str(name), str(value),
                  MimeTypeExtractor.RESULT_SUCCESS if passed else MimeTypeExtractor.RESULT_FAIL, str(url))


    # variable for result
    ret = True
    try:
      if not isinstance(crawledResource, CrawledResource):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(crawledResource)), str(type(CrawledResource()))))

      if not isinstance(propertyDict, dict):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(propertyDict)), str(type(dict()))))

      if not isinstance(urlObj, dc_event.URL):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(urlObj)), str(type(dc_event.URL(None, None)))))

      # check regular expression for domain
      if MimeTypeExtractor.OPTION_URL_EXPRESSION in propertyDict and isinstance(urlObj.url, basestring) and \
        re.search(pattern=propertyDict[MimeTypeExtractor.OPTION_URL_EXPRESSION], string=urlObj.url, flags=re.U + re.I) is None:
        ret = False
        printStatus(name=MimeTypeExtractor.OPTION_URL_EXPRESSION, value=propertyDict[MimeTypeExtractor.OPTION_URL_EXPRESSION],
                    url=urlObj.url, log=log, passed=ret)

      # check URLs type
      if ret and MimeTypeExtractor.OPTION_URL_TYPES in propertyDict:
        if not isinstance(propertyDict[MimeTypeExtractor.OPTION_URL_TYPES], list):
          raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(propertyDict[MimeTypeExtractor.OPTION_URL_TYPES])), str(type(list()))))

        if len(propertyDict[MimeTypeExtractor.OPTION_URL_TYPES]) == 0:
          ret = True
        else:
          ret = urlObj.type in propertyDict[MimeTypeExtractor.OPTION_URL_TYPES]
        printStatus(name=MimeTypeExtractor.OPTION_URL_TYPES, value=propertyDict[MimeTypeExtractor.OPTION_URL_TYPES], url=urlObj.url, log=log, passed=ret)

      # check URLs parent
      if ret and MimeTypeExtractor.OPTION_URL_PARENT in propertyDict:
        if not isinstance(propertyDict[MimeTypeExtractor.OPTION_URL_PARENT], list):
          raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(propertyDict[MimeTypeExtractor.OPTION_URL_PARENT])), str(type(list()))))

        if (urlObj.parentMd5 and MimeTypeExtractor.URL_PARENT_TYPE_VALUE_ONLY_NOT_PARENT not in propertyDict[MimeTypeExtractor.OPTION_URL_PARENT]) or \
          (not urlObj.parentMd5 and MimeTypeExtractor.URL_PARENT_TYPE_VALUE_ONLY_PARENT not in propertyDict[MimeTypeExtractor.OPTION_URL_PARENT]):
          ret = False
        printStatus(name=MimeTypeExtractor.OPTION_URL_PARENT, value=propertyDict[MimeTypeExtractor.OPTION_URL_PARENT], url=urlObj.url, log=log, passed=ret)

      if ret:
        # extract content-types
        contentTypes = []
        if MimeTypeExtractor.OPTION_CONTENT_TYPES in propertyDict:
          if not isinstance(propertyDict[MimeTypeExtractor.OPTION_CONTENT_TYPES], list):
            raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(propertyDict[MimeTypeExtractor.OPTION_CONTENT_TYPES])), str(type(list()))))

          contentTypes = propertyDict[MimeTypeExtractor.OPTION_CONTENT_TYPES]

        if log is not None:
          log.debug("crawledResource.content_type = '%s'", str(crawledResource.content_type))

        # extract mode
        mode = int(propertyDict[MimeTypeExtractor.OPTION_MODE]) if MimeTypeExtractor.OPTION_MODE in propertyDict else MimeTypeExtractor.DEFAULT_VALUE_MODE

        if mode == MimeTypeExtractor.MODE_VALUE_ALWAYS:
          # nothing to do
          pass
        elif mode == MimeTypeExtractor.MODE_VALUE_NOT_DEFINED_FROM_HTTP_HEADER_RESPONSE:
          # check not defined content-type from HTTP header response
          ret = crawledResource.content_type == ""
        elif mode == MimeTypeExtractor.MODE_VALUE_MATCH_CONTENT_TYPES:
          # check of the match in content types list
          ret = crawledResource.content_type in contentTypes
        elif mode == MimeTypeExtractor.MODE_VALUE_NOT_MATCH_CONTENT_TYPES:
          # check of the not match in content types list
          ret = crawledResource.content_type not in contentTypes

        printStatus(name=MimeTypeExtractor.OPTION_MODE, value=mode, url=urlObj.url, log=log, passed=ret)
        printStatus(name=MimeTypeExtractor.OPTION_CONTENT_TYPES, value=contentTypes, url=urlObj.url, log=log, passed=ret)

    except Exception, err:
      if log is not None:
        log.error(str(err))
        log.info("Used propertyDict: %s", str(propertyDict))

    return ret


  # # Check is allowed for detection mime type
  #
  # @param crawledResource - crawler resource instance
  # @param properties - properties string or dictionary
  # @param urlObj - URLs object instance
  # @param log - logger instance
  # @return True if allowed or otherwise False
  @staticmethod
  def isAllowedDetect(crawledResource, properties, urlObj, log=None):
    # variable for result
    ret = True
    try:
      if not isinstance(crawledResource, CrawledResource):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(crawledResource)), str(type(CrawledResource()))))

      if not isinstance(urlObj, dc_event.URL):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(urlObj)), str(type(dc_event.URL(None, None)))))

      propertyObj = Utils.jsonLoadsSafe(jsonString=properties, default=property, log=log)
      if isinstance(propertyObj, list):
        results = []
        for propertyDict in propertyObj:
          results.append(MimeTypeExtractor.isAllowedRuleElement(crawledResource, propertyDict, urlObj, log))

        if log is not None:
          log.info("results: %s", str(results))

        ret = (True in results) if len(results) > 0 else True
      elif isinstance(propertyObj, dict) :
        ret = MimeTypeExtractor.isAllowedRuleElement(crawledResource, propertyObj, urlObj, log)

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  ## extract mime type
  #
  # @param crawledResource - crawler resource instance
  # @param siteProperties - site properties
  # @param urlObj - URLs object instance
  # @param log - logger instance
  # @param default - default value for result
  # @return detected mime type
  @staticmethod
  def extract(crawledResource, siteProperties, urlObj, log=None, default=None):
    # variable for result
    ret = default
    try:
      if not isinstance(siteProperties, dict):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(siteProperties)), str(type(dict()))))

      if APP_CONSTS.MIME_TYPE_AUTO_DETECT in siteProperties and \
        MimeTypeExtractor.isAllowedDetect(crawledResource=crawledResource,
                                          properties=siteProperties[APP_CONSTS.MIME_TYPE_AUTO_DETECT],
                                          urlObj=urlObj,
                                          log=log):
        # extract content type map
        contentTypeMap = None
        if APP_CONSTS.CONTENT_TYPE_MAP in siteProperties:
          contentTypeMap = Utils.jsonLoadsSafe(jsonString=siteProperties[APP_CONSTS.CONTENT_TYPE_MAP],
                                               default=siteProperties[APP_CONSTS.CONTENT_TYPE_MAP],
                                               log=log)
        # mime type detection
        ret = MimeTypeExtractor.detect(crawledResource=crawledResource, contentTypeMap=contentTypeMap, log=log)

    except Exception, err:
      if log is not None:
        log.error(str(err))
        log.info("Used siteProperties: %s", str(siteProperties))

    return ret


  # # detect mime type for URL by HEAD method
  #
  # @param url - url string
  # @param contentTypeMap - content type map
  # @param log - logger instance
  # @return mime type string or None
  @staticmethod
  def detectUrlMime(url, contentTypeMap=None, log=None):
    # variable for result
    ret = None
    try:
      res = requests.head(url, timeout=MimeTypeExtractor.DEFAULT_VALUE_HEAD_TIMEOUT)
      ret = res.headers.get(MimeTypeExtractor.HEADERS_CONTENT_TYPE_OPTIONS_NAME, '').split(';')[0]
      if log is not None:
        log.debug("Extracted by HEAD '%s' = '%s'", str(MimeTypeExtractor.HEADERS_CONTENT_TYPE_OPTIONS_NAME), str(ret))

      if isinstance(contentTypeMap, dict) and ret in contentTypeMap:
        if log is not None:
          log.debug("Mime type replaced from '%s' to '%s'", str(ret), str(contentTypeMap[ret]))

        ret = str(contentTypeMap[ret])

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  ## extract mime type for URL by HEAD method
  #
  # @param url - url string
  # @param siteProperties - site properties
  # @param log - logger instance
  # @param default - default value for result
  # @return extracted mime type
  @staticmethod
  def extractUrlMime(url, siteProperties, log=None, default=None):
    # variable for result
    ret = default
    try:
      if not isinstance(siteProperties, dict):
        raise Exception(MimeTypeExtractor.ERROR_INPUT_TYPE % (str(type(siteProperties)), str(type(dict()))))

      # extract content type map
      contentTypeMap = None
      if APP_CONSTS.CONTENT_TYPE_MAP in siteProperties:
        contentTypeMap = Utils.jsonLoadsSafe(jsonString=siteProperties[APP_CONSTS.CONTENT_TYPE_MAP],
                                             default=siteProperties[APP_CONSTS.CONTENT_TYPE_MAP],
                                             log=log)
      # detect mime type for URL
      ret = MimeTypeExtractor.detectUrlMime(url=url, contentTypeMap=contentTypeMap, log=log)

    except Exception, err:
      if log is not None:
        log.error(str(err))
        log.info("Used siteProperties: %s", str(siteProperties))

    return ret
