"""
HCE project, Python bindings, Distributed Tasks Manager application.
Class RefererHeaderResolver content main functional for resolve 'Referer' value for header

@package: dc_crawler
@file RefererHeaderResolver.py
@author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
@copyright: Copyright &copy; 2013-2019 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

import re

import dc.EventObjects
import app.Utils as Utils
import app.Consts as APP_CONSTS


class RefererHeaderResolver(object):
  # # Constants used in class
  MODE_NOT_ADD = 0
  MODE_ADD_URL = 1
  MODE_ADD_DOMAIN = 2
  MODE_ADD_PARENT_URL = 3

  DEFAULT_VALUE_MODE = MODE_NOT_ADD

  REFERER_HEADER_NAME = "Referer"
  PATTERN_ALLOWED_ALL = '*'


  # # Initialization
  def __init__(self):
    pass


  # # Get URL from DB by MD5
  #
  # @param siteId - current resource's siteId
  # @param urlMd5 - current resource's URL MD5
  # @param dbWrapper - DBWrapper instance
  # @param log - logger instance
  # @return resolved URL or None
  @staticmethod
  def getUrlByMD5(siteId, urlMd5, dbWrapper, log=None):
    ret = None
    try:
      if siteId is not None and urlMd5 is not None and dbWrapper is not None:
        urlStatusObj = dc.EventObjects.URLStatus(siteId, urlMd5)
        result = dbWrapper.urlStatus(urlStatusObj, True)
        if log is not None:
          log.info("get URL by MD5 result: %s", Utils.varDump(result))

        if len(result) > 0 and isinstance(result[0], dc.EventObjects.URL):
          ret = result[0].url

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  ## Get parent URL from DB
  #
  # @param siteId - current resource's siteId
  # @param url - current resource's parentMd5
  # @param dbWrapper - DBWrapper instance
  # @param log - logger instance
  # @return parent URL or None
  @staticmethod
  def getParentUrl(siteId, url, dbWrapper, log=None):
    ret = None
    try:
      if siteId is not None and url is not None and dbWrapper is not None:
        # get parent MD5
        urlStatusObj = dc.EventObjects.URLStatus(siteId, url)
        result = dbWrapper.urlStatus(urlStatusObj)
        if log is not None:
          log.info("get parent MD5 result: %s", Utils.varDump(result))

        if len(result) > 0 and isinstance(result[0], dc.EventObjects.URL) and result[0].parentMd5 != "":
          # get parent URL use parent MD5
          ret = RefererHeaderResolver.getUrlByMD5(siteId, result[0].parentMd5, dbWrapper, log)

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  ## Is allowed URL
  #
  # @param pattern - pattern string
  # @param url - URL string
  # @param log -logger instance
  @staticmethod
  def isAllowedURL(pattern, url, log=None):
    #variable for result
    ret = False
    try:
      if isinstance(pattern, basestring) and isinstance(url, basestring):
        if pattern.strip() == RefererHeaderResolver.PATTERN_ALLOWED_ALL:
          ret = True
        else:
          if re.search(pattern, url, flags=re.U + re.I) is not None:
            ret = True

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  # # Get mode number
  #
  # @param propertyValue - property value
  # @param url - current resource's URL
  # @param log - logger instance
  # @return mode number
  @staticmethod
  def getModeNumber(propertyValue, url, log=None):
    # variable for result
    ret = RefererHeaderResolver.DEFAULT_VALUE_MODE
    try:
      value = Utils.jsonLoadsSafe(jsonString=propertyValue, default=propertyValue)
      if isinstance(value, int):
        ret = value
      elif isinstance(value, list):
        for elem in value:
          if isinstance(elem, list):
            if len(elem) > 1:
              if isinstance(elem[0], basestring) and RefererHeaderResolver.isAllowedURL(pattern=elem[0], url=url, log=log):
                ret = int(elem[1])
                break
              elif isinstance(elem[0], list):
                found = False
                for elemPattern in elem[0]:
                  if isinstance(elemPattern, basestring) and RefererHeaderResolver.isAllowedURL(pattern=elemPattern, url=url, log=log):
                    ret = int(elem[1])
                    found = True
                    break

                if found:
                  break

    except Exception, err:
      if log is not None:
        log.error(str(err))

    return ret


  # # Update headers by resolved 'Referer'
  #
  # @param siteProperties - site properties
  # @param headers - HTTP headers dictionary
  # @param url - current resource's URL
  # @param siteId  - current resource's siteId
  # @param dbWrapper - DBWrapper instance
  # @param log - logger instance
  # @return - None
  @staticmethod
  def updateHeaders(siteProperties, headers, url, siteId, dbWrapper=None, log=None):
    if APP_CONSTS.REFERER_SELF_URL in siteProperties and isinstance(headers, dict):
      if log is not None:
        log.debug("Found property '%s'", str(APP_CONSTS.REFERER_SELF_URL))

      # check already exists 'Referer'
      isExists = False
      for headerName in headers:
        if headerName.lower() == RefererHeaderResolver.REFERER_HEADER_NAME.lower():
          isExists = True

      if log is not None:
        log.debug("'%s' already exists = %s", str(RefererHeaderResolver.REFERER_HEADER_NAME), str(isExists))

      if not isExists:
        mode = RefererHeaderResolver.getModeNumber(propertyValue=siteProperties[APP_CONSTS.REFERER_SELF_URL], url=url, log=log)
        if log is not None:
          log.debug("Extracted mode = %s", str(mode))

        # apply different algorithms
        if mode == RefererHeaderResolver.MODE_NOT_ADD:
          pass
        elif mode == RefererHeaderResolver.MODE_ADD_URL:
          headers[RefererHeaderResolver.REFERER_HEADER_NAME] = url
        elif mode == RefererHeaderResolver.MODE_ADD_DOMAIN:
          headers[RefererHeaderResolver.REFERER_HEADER_NAME] = Utils.UrlParser.generateDomainUrl(url)
        elif mode == RefererHeaderResolver.MODE_ADD_PARENT_URL:
          parentUrl = RefererHeaderResolver.getParentUrl(siteId, url, dbWrapper, log)
          headers[RefererHeaderResolver.REFERER_HEADER_NAME] = url if parentUrl is None else parentUrl
